#ifndef ASM_E2K_ASSEMBLER_INL_H
#define ASM_E2K_ASSEMBLER_INL_H

#include "registers.h"
#include "assembler-e2k.h"

namespace asm_e2k {

const int Assembler::_c1f1_shift[] = { 21, -1, 14, 7, -1, 0 };


PredDefMap::PredDefMap() : first_set(SizeInBits - 1), last_set(0) {
    std::fill(data, data + SizeInWords, 0);
}

void PredDefMap::set(uint32_t idx) {
    first_set = std::min(first_set, idx);
    last_set  = std::max(last_set, idx);
    data[word_index(idx)] |= bit_mask(idx);
}

bool PredDefMap::get(uint32_t idx) {
    return data[word_index(idx)] & bit_mask(idx);
}

void PredDefMap::reset() {
    uint32_t first_set_word = word_index(first_set);
    uint32_t last_set_word = word_index(last_set);
    for(uint32_t i = first_set_word; i <= last_set_word; i++) {
        data[i] = 0;
    }
#ifdef ASM_E2K_DEBUG
    for(uint32_t i = 0; i < SizeInWords; i++) {
        ASM_E2K_ASSERT(data[i]==0);

    }
#endif
    first_set = SizeInBits - 1;
    last_set = 0;
}

void PredDefMap::set_pred_def(const uint32_t def_idx) {
    set(def_idx);
}

void PredDefMap::set_pred_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse) {
    int32_t pr_idx;

    ASM_E2K_ASSERT(pred_idx != EREG_PRX || inverse == false);
    if (pred_idx == EREG_PRX) {
        pr_idx = def_idx;
    } else {
        uint32_t pred_idx_ = pred_idx - EREG_PR0;
        ASM_E2K_ASSERT(pred_idx_ >=0 && pred_idx_ < PredicatesUsedCount);
        pr_idx = TotalRegCount * (pred_idx_ * 2 + inverse + 1) + def_idx;
    }
    ASM_E2K_ASSERT(pr_idx <= SizeInBits);
    set(pr_idx);
}

bool PredDefMap::get_pred_def(const uint32_t def_idx) {
    return get(def_idx);
}

bool PredDefMap::get_pred_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse) {
    int32_t pr_idx;
    ASM_E2K_ASSERT(pred_idx != EREG_PRX || inverse == false);
    if (pred_idx == EREG_PRX) {
        pr_idx = def_idx;
    } else {
        uint32_t pred_idx_ = pred_idx - EREG_PR0;
        ASM_E2K_ASSERT(pred_idx_ >=0 && pred_idx_ < PredicatesUsedCount);
        pr_idx = TotalRegCount * (pred_idx_ * 2 + inverse + 1) + def_idx;
    }
    return get(pr_idx);
}

void AsmScheduler::init_scheduler(Assembler *a, int inst_cnt) {

    insts_cnt = inst_cnt;
    assm = a;

    defs.clear();
    invaruses.clear();
    temp_regs.clear();
    if (curr_num != 0) {
       insts_cnt = curr_num;
    }
#ifdef ASM_SCHED_DEBUG
    if (debug_flags & shed_debug_build) {
        printf("init_scheduler clear %d elements curr_num = %d size = %d\n",
            insts_cnt, curr_num,sizeof (struct SchedInstruction) * insts_cnt );
    }
#endif
    curr_num = 0;
    max_depth = 0;
    last_load = 0;
    last_store = 0;
}

#ifdef ASM_SCHED_DEBUG
static int cmpile_cnt = 0;
#endif

void Assembler::init_asm_sheduler(int level, bool sched_debug, bool need_patch, int free_reg, int free_reg_cnt, int insts_cnt) {
    sched_level = level;
    if (!sched) {
        asm_sched = new (&sched_buf) AsmScheduler;
        insts_cnt = MAX_INSTRUCTS_IN_BLOCK;

 #ifdef ASM_SCHED_DEBUG
     char *flag = getenv("ASM_SCHED_DEBUG");
     if(flag) {
         asm_sched->debug_flags = atoi(flag);
 //          printf("ASM_SCHED_DEBUG  = %d\n", asm_sched->debug_flags);
     }
     if (asm_sched->debug_flags & shed_debug_build)
         printf("new (&sched_buf) AsmScheduler sched_debug:%d &sched_buf: %x\n",
                  sched_debug, &sched_buf);
#endif
    }
    patch_offsets = need_patch;
    sched = asm_sched;
    sched->debug = sched_debug;
    sched->free_register = free_reg;
    sched->free_registers_cnt = free_reg_cnt;
    if (patch_offsets) {
      sched->rbs_offsets.clear();
      sched->wsz_offsets.clear();
    }
    sched->init_scheduler(this, insts_cnt);
#ifdef ASM_SCHED_DEBUG
    if (sched && (sched->debug_flags & shed_debug_build || sched->debug_flags & shed_debug_patching ||
                  sched->debug_flags & shed_debug_input || sched->debug_flags & shed_debug_output)) {
        printf("init_asm_sheduler: cmpile_cnt=%d free_reg =%d, free_reg_cnt =%d\n",cmpile_cnt, free_reg, free_reg_cnt);
        printf("_bundle_state == Open : %d,  is_empty_bundle() = %d\n",  _bundle_state == Open, is_empty_bundle());
    }
#endif
    if (is_empty_bundle()) {
        _bundle_state = Close;
    }
#ifdef ASM_SCHED_DEBUG
    if (sched->debug_flags & shed_debug_build)
        printf("finish init_asm_sheduler: cmpile_cnt=%d\n",cmpile_cnt);
#endif
}

#define CS1_RBS_VALUE_OFF 6
#define CS1_RBS_BITSIZE   6
#define CS1_WSZ_VALUE_OFF 12
#define CS1_WSZ_BITSIZE   5
#define CS1_WSZ_REMAIND   32

void Assembler::e2k_asm_patch_rbs(int32_t *addr, int32_t new_rbs) {
    uint32_t prev_rbs = *addr;
    *addr = (prev_rbs >> CS1_RBS_VALUE_OFF << CS1_RBS_BITSIZE) | new_rbs;
}

void Assembler::e2k_asm_patch_wsz(int32_t *addr, int32_t new_wsz) {
    uint32_t bits12 = *addr >> CS1_WSZ_VALUE_OFF;
    uint32_t bits5 = *addr % CS1_WSZ_REMAIND;
    uint32_t new_wd = new_wsz;
    *addr = bits12 << CS1_WSZ_VALUE_OFF | new_wd << CS1_WSZ_BITSIZE | bits5;
}

void Assembler::e2k_print_asm_info() {
    int count1 = sched->rbs_offsets.size();
    if (sched && count1 > 0) {
        for (auto off : sched->rbs_offsets) {
            uint32_t *addr = reinterpret_cast<uint32_t*>(buffer_ + off);
#ifdef ASM_SCHED_DEBUG
            if (sched && (sched->debug_flags & shed_debug_build)) {
                printf("e2k_print_asm_info pc_offset();  %d off: %d ((*addr)>>28)%4: %d addr: %ld buffer_: %ld\n",
                       pc_offset(), off, ((*addr)>>28)%4, addr, buffer_);
            }
#endif
        }
    }
}

void Assembler::finish_asm_sheduler() {
#ifdef ASM_SCHED_DEBUG
    if (sched && sched->debug_flags & shed_debug_build)
         printf("finish_asm_sheduler: cmpile_cnt=%d is_empty_bundle() = %d patch_offsets : %d\n",
               cmpile_cnt, is_empty_bundle(), patch_offsets);
#endif

     if (sched_level > 0 &&  patch_offsets) {
         int new_wbs = (sched->free_register - EREG_R0 + 1 + sched->free_registers_cnt ) / 2;
         int rmap_wbs = wsz_to_wbs(new_wbs);
         rmap_wbs = rmap_wbs < 4 ? 4 : rmap_wbs;

         int count = sched->rbs_offsets.size();
 #ifdef ASM_SCHED_DEBUG
         if (sched && (sched->debug_flags & shed_debug_patching) ) {
              printf("finish_asm_sheduler: new_wbs=%d rmap_wbs = %d \n",
                     new_wbs, rmap_wbs); // wsz_to_wbs(rmap_wbs)
              printf("finish_asm_sheduler: sched->free_register=%d sched->free_registers_cnt = %d EREG_R0 = %d\n",
                     sched->free_register, sched->free_registers_cnt, EREG_R0);
         }
#endif
         if (count > 0) {
             for (auto off : sched->rbs_offsets) {
#ifdef V8_TARGET_ARCH_E2K
                int32_t *addr = reinterpret_cast<int32_t*>(*buffer_ + off);
#else
                int32_t *addr = reinterpret_cast<int32_t*>(buffer_ + off);
#endif
#ifdef ASM_SCHED_DEBUG
                if (sched && (sched->debug_flags & shed_debug_patching)) {
                    printf("finish_asm_sheduler off: %x buffer_: %ld\n",off, buffer_);
                }
                if (sched && (sched->debug_flags & shed_debug_patching) ) {
                     printf("finish_asm_sheduler: addr=%ld off = %x buffer_: %ld\n",
                            addr, off, buffer_);
                     printf("old *addr =%ld old rbs =%d \n", addr, (*addr)%64 );
                }
 #endif
                e2k_asm_patch_rbs(addr, rmap_wbs);
#ifdef ASM_SCHED_DEBUG
                if (sched && (sched->debug_flags & shed_debug_patching) ) {
                     printf("new *addr =%ld new rbs =%d \n", addr, (*addr)%64 );
                }
 #endif
             }
         }
         count = sched->wsz_offsets.size();
         if (count > 0) {
             for (auto off : sched->wsz_offsets) {
#ifdef V8_TARGET_ARCH_E2K
                int32_t *addr = reinterpret_cast<int32_t*>(*buffer_ + off);
#else
                int32_t *addr = reinterpret_cast<int32_t*>(buffer_ + off);
#endif
#ifdef ASM_SCHED_DEBUG
                if (sched && (sched->debug_flags & shed_debug_patching) ) {
                     printf("finish_asm_sheduler: addr=%ld off = %x buffer_: %ld\n",
                            addr, off, buffer_);
                }
#endif
#ifdef ASM_SCHED_DEBUG
                if (sched && (sched->debug_flags & shed_debug_patching) ) {
                    uint32_t old_wd = (*(addr) >> 5) % 128;
                    printf("setwd addr = %ld old_wd =%d\n", addr, old_wd);
                }
#endif
                e2k_asm_patch_wsz(addr, new_wbs);
#ifdef ASM_SCHED_DEBUG
                if (sched && (sched->debug_flags & shed_debug_patching) ) {
                    uint32_t new_wd = (*(addr) >> 5) % 128;
                    printf("new setwd addr = %ld new_wd =%d\n", addr, new_wd);
                }
#endif
             }
         }
     }
     sched_level = 0;
}

Assembler::Assembler() : bw(this), defs(), sched_level(0)
{
    lts_num = 0;
    buffer_size_ = 2 * KB;
    buffer_ = new byte[buffer_size_];
    is_external_buffer = false;
    pc_ = buffer_;
    prev_pc_offset = -1;
    _bundle_state = Close;
}

Assembler::Assembler(byte* buffer, unsigned offset, unsigned buffer_size) : bw(this), buffer_(buffer),
    buffer_size_(buffer_size), is_external_buffer(true), sched_level(0), defs()
{
    ASM_E2K_ASSERT(offset < buffer_size);
    lts_num = 0;
    pc_ = buffer + offset;
    prev_pc_offset = -1;
    _bundle_state = Close;
}

void Assembler::set_buffer(byte* buffer, unsigned offset, unsigned buffer_size) {
    ASM_E2K_ASSERT(is_external_buffer);
    ASM_E2K_ASSERT(offset < buffer_size);
    buffer_ = buffer;
    pc_ = buffer + offset;
    buffer_size_ = buffer_size;
}

Assembler::~Assembler()
{
    if(!is_external_buffer) {
      delete [] buffer_;
    }
}

} // namespace asm_e2k

#endif // ASM_E2K_ASSEMBLER_INL_H

