#ifndef ASM_E2K_ASSEMBLER_H
#define ASM_E2K_ASSEMBLER_H

#include "macro.h"

NAMESPACE_ASM_E2K_BEGIN

enum DataTypes {
  DATA_U1   = 0,
  DATA_I8   = 1,
  DATA_U8   = 2,
  DATA_I16  = 3,
  DATA_U16  = 4,
  DATA_I32  = 5,
  DATA_U32  = 6,
  DATA_I64  = 7,
  DATA_U64  = 8,
  DATA_S    = 9,
  DATA_D    = 11,
  DATA_X    = 13,
  DATA_LAST = 14
};

NAMESPACE_ASM_E2K_END

#include "registers.h"
#include "operands.h"
#include "label.h"
#include "opcodes-e2k.h"
#include "latency.h"

#ifdef __cplusplus

#include <array>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <memory.h>
#include <map>
#include <list>
#include <iostream>
typedef std::map <uint64_t, int> RegMap;
typedef std::multimap <uint64_t, int> RegMultiMap;
typedef std::pair <uint64_t, int> IntPair;
typedef std::pair <int, uint16_t> IntShortPair;
typedef std::map <int, uint16_t> IntShortMap;
typedef std::list <int> RegList;
typedef std::list <long> LongList;

typedef uint8_t byte;
typedef byte* address;
typedef uint32_t AlsOffset;

const int KB = 1024;
const int MB = KB * KB;

NAMESPACE_ASM_E2K_BEGIN

extern const ALUInfo alu_info[];

#define ILLEGAL_ALS_ID -1 // calculating als id in Bundle

#define REG_PRX_NUM  (EREG_PRX - EREG_PR0)

#define SETWD_LTS_SIZE 4

#define MAX_LTS_SIZE 16 // max summ of literal syllables size

#define MAX_INSTRUCTS_IN_BLOCK 600

#define shed_debug_build   1
#define shed_debug_sched   2
#define shed_debug_latency 4
#define shed_debug_patching 8
#define shed_debug_input   16
#define shed_debug_output  32
#define shed_debug_curr    64

enum LtsType {
    LTS_SPECIAL,
    LTS_16,
    LTS_32,
    LTS_64,
    LTS_LAST
};

class Lts {
public:
    Lts() {}
    Lts(LtsType type, uint64_t val, int32_t als_idx, Relocation reloc);
    Lts(const Lts &other) : Lts(other.type, other.val, other.als_idx, other.reloc) { }

public:
    LtsType type;    // 16, 32, 64 or special LTS0
    uint64_t val;    // lts value
    int8_t als_idx;
    uint8_t size;    // tls size in bytes
    Relocation reloc;
};

class Assembler;
class PredDefMap;

// -----------------------------------------------------------------------------
// Machine instruction Operands

enum ScaleFactor {
    times_1 = 0,
    times_2 = 1,
    times_4 = 2,
    times_8 = 3,
    times_int_size = times_4,
    times_pointer_size = times_8
};

// Contains all the definitions needed for Elbrus assembly code generation.
// The Elbrus Assembler: Pure assembler doing NO optimizations on the instruction
// level (e.g. mov rax, 0 is not translated into xor rax, rax!); i.e., what you write
// is what you get. The Assembler is generating code into a buffer.
class BitWriter {
public:
    BitWriter(Assembler* _asm) : _asm(*_asm), buff(0), flag(false) {}
    void            init();
    void            flush();
    void            half(uint16_t);

public:
    Assembler&      _asm;
    uint32_t        buff;
    bool            flag;
};

class Instruction{
    explicit Instruction(uint32_t* inst_) : inst(inst_) {}

    uint32_t size() {
        uint32_t hs = *reinterpret_cast<uint32_t*>(inst);
        uint32_t size = ((hs & 0x00000070) >> 4) + 1;
        return size;
    }

private:
    uint32_t* inst;

};

class PredDefMap {
private:
    enum {
        BitsPerWord = 64,
        BytesPerWord = 8,
        LogBitsPerWord = 6,
        SizeInBits = PredicateThreadsNum * TotalRegCount,
        SizeInWords = (SizeInBits >> LogBitsPerWord) + 1  
    };

    uint64_t data[SizeInWords];
    uint32_t first_set;
    uint32_t last_set;

    inline static uint32_t bit_in_word(int bit) { return bit & (BitsPerWord - 1); }
    inline static uint32_t word_index(int bit)  { return bit >> LogBitsPerWord; }
    inline static uint64_t bit_mask(int bit) { return (uint64_t)1 << bit_in_word(bit); }
    inline void set(uint32_t idx);
    inline bool get(uint32_t idx);
public:
    PredDefMap();

    void set_pred_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse);
    void set_pred_def(const uint32_t def_idx) ;
    bool get_pred_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse);
    bool get_pred_def(const uint32_t def_idx) ;
    void reset();
};

static int wsz_to_wbs(int wsz) {
    return ((wsz * 2) - 8) / 2;
}

class AsmScheduler {
 public:
    enum Coding {
        ALOPF1,
        ALOPF_LD,
        ALOPF_LD_MAS,
        ALOPF2,
        ALOPF2ext,
        ALOPF3,
        ALOPF_3_MAS,
        ALOPF7,
        ALOPF8,
        ALOPF11,
        ALOPF12,
        ALOPF15,
        ALOPF16,
        MERGE,
        COPF2,
        COPF4encode,
        COPF4,
        COPF4extEncode,
        COPF4ext,
        CALL,
        IBRANCH,
        SETBN,
        CT,
        PLOPF
    };
    struct SchedInstruction {
        Coding code;
        E2kOpCode opc;
        uint32_t opce; // cmpopce, opc for PLOPF
        int64_t pr_idx; // pr_idx
        int64_t wbs_val;
        int64_t use1_idx;    // use_idx
        int64_t use2_idx;
        int64_t use3_idx;  // m_pr_idx
        int64_t def_idx;   // dst_idx
        bool fixed_size = false; // LTS should not try to optimize its size
        int use1_reloc;      // use_reloc
        int use2_reloc;
        DataTypes use1_type; // use_type
        DataTypes use2_type;
        uint32_t ctp_opc;
        unsigned disp;
        uint32_t uopc;
        uint32_t opx;
        unsigned udisp;
        unsigned param;
        unsigned type;
        unsigned rpsz;
        unsigned wsz;
        unsigned wbs;
        unsigned nfx;
        unsigned dbl;
        uint32_t mas;
        uint32_t opc2;
        uint32_t opce2;
        uint32_t reg_code;
        uint32_t neg0;
        uint32_t neg1;
        uint32_t vdst;

        bool use1_reg;  // use_reg
        bool use2_reg;
        bool use3_reg;
        bool cinverse = false;
        bool inverse = false; // mininverce;
        bool speculative = false;
        bool visited = false;
        bool last_in_cycle = false;
        bool is_induction = false;
    } instructs[MAX_INSTRUCTS_IN_BLOCK];
    struct InstrctionAttrs {
        RegList uses;
        RegList defs;
        int depth;
        int max_src_depth;
    } attrs[MAX_INSTRUCTS_IN_BLOCK];
 public:
    Assembler* assm;
    RegMap defs;
    RegMap temp_regs;
    RegMultiMap invaruses;
    LongList rbs_offsets;
    LongList wsz_offsets;
    uint32_t insts_cnt;
    uint32_t curr_num;
    uint32_t nops;
    uint32_t last_load;   // (set inst + 1) to use 0 for check
    uint32_t last_store;  // (set inst + 1) to use 0 for check
    uint32_t max_depth;
    int32_t free_register;
    int32_t free_registers_cnt;
    uint8_t sched_level = 0;
#ifdef ASM_SCHED_DEBUG
    uint8_t debug_flags = 0;
#endif
    bool is_setwd = false;
    bool debug;
public:
    AsmScheduler() {};
    AsmScheduler(Assembler* a) {assm = a; };
    ~AsmScheduler() {};
    void init_scheduler(Assembler *a, int insts_cnt);
    void do_schedule();
    void init_instr(int inst);
    void inst_process(int inst);
    bool has_def(int inst);
    void use_process(int inst, int64_t use_idx);
    int  get_bypass_lat(int inst, uint16_t exec_unit);
    bool check_valid_distance(int inst);
    void def_process(int inst, int64_t def_idx);
    int get_temp_register(int src_inst, int trg_inst);
    int can_be_add_to_bundle(int inst);
    int try_alu(int inst);
};


class Assembler
{
public:
    enum Condition {
        Zero                = 0x4,
        NonZero             = 0x5,
        Equal               = 0x4,
        NotEqual            = 0x5,
        LessThan            = 0xc,
        LessThanOrEqual     = 0xe,
        GreaterThan         = 0xf,
        GreaterThanOrEqual  = 0xd,
        Below               = 0x2,
        BelowOrEqual        = 0x6,
        Above               = 0x7,
        AboveOrEqual        = 0x3,
        Overflow            = 0x0,
        NoOverflow          = 0x1,
        CarrySet            = 0x2,
        CarryClear          = 0x3,
        Negative            = 0x8,
        Positive            = 0x9,
        Parity              = 0xa,
        NoParity            = 0xb,
        Signed              = 0x10,
        NotSigned           = 0x11,
        AndZeroEqual
    };

    enum DoubleCondition {
        // These conditions will only evaluate to true if the comparison is ordered - i.e. neither operand is NaN.
        DoubleOrdered,
        DoubleEqual,
        DoubleNotEqual,
        DoubleGreaterThan,
        DoubleGreaterThanOrEqual,
        DoubleLessThan,
        DoubleLessThanOrEqual,
        DoubleUnordered,
        DoubleEqualOrUnordered,
        DoubleNotEqualOrUnordered,
        DoubleGreaterThanOrUnordered,
        DoubleGreaterThanOrEqualOrUnordered,
        DoubleLessThanOrUnordered,
        DoubleLessThanOrEqualOrUnordered
    };

    enum NaNCond {
        NaN_HandledByCond,
        NaN_IsTrue,
        NaN_IsFalse
    };

    // If the primary condition returned by ConditionFromDoubleCondition doesn't
    // handle NaNs properly, return NaN_IsFalse if the comparison should be
    // overridden to return false on NaN, NaN_IsTrue if it should be overridden
    // to return true on NaN, or NaN_HandledByCond if no secondary check is
    // needed.
    static inline NaNCond NaNCondFromDoubleCondition(DoubleCondition cond) {
        switch (cond) {
        case DoubleOrdered:
        case DoubleNotEqual:
        case DoubleGreaterThan:
        case DoubleGreaterThanOrEqual:
        case DoubleLessThan:
        case DoubleLessThanOrEqual:
        case DoubleUnordered:
        case DoubleEqualOrUnordered:
        case DoubleGreaterThanOrUnordered:
        case DoubleGreaterThanOrEqualOrUnordered:
        case DoubleLessThanOrUnordered:
        case DoubleLessThanOrEqualOrUnordered:
            return NaN_HandledByCond;
        case DoubleEqual:
            return NaN_IsFalse;
        case DoubleNotEqualOrUnordered:
            return NaN_IsTrue;
        }
        ASM_E2K_CRASH("Unknown double condition");

    }

    // header fields
    uint32_t         hs_al;
    uint32_t         hs_ale;
    uint32_t         hs_pl;
    uint32_t         hs_cd_;  // num of rlp operations
    uint32_t         hs_c;
    uint32_t         hs_sw;
    uint32_t         hs_s;
    uint32_t         hs_x;
    uint32_t         hs_lm;
    uint32_t         hs_nop;
    uint32_t         hs_lng_; // in half words
    uint32_t         hs_mdl_; // in half words

    // bundle fields
    uint32_t         hs;
    uint32_t         ss;
    uint32_t         als[6];
    uint32_t         cs[2];
    uint32_t         pad;
    uint16_t         ale[6];
    uint16_t         aas[6];
    std::array<Lts, 6> ltss;
    uint8_t          lts_num;
    uint32_t         pls[3];
    uint16_t         cds[6];

    int32_t          lts_len;       // len of lts area in bytes
    int32_t          lts16_num;     // num allocated slots for lts16 (can't be > 4)
    int32_t          c0_offset;     // c0 syllable offset in bundle

    static bool zero_bits_sup; // old procs not support zero bits
    int sched_level;
    bool asm_statistic = false;
    bool patch_offsets = false;
    uint32_t bundle_cnt;
    uint32_t nop_cnt;
    uint32_t inst_cnt;
#ifdef ASM_E2K_DEBUG
    bool is_disp;
#endif
    AsmScheduler* sched = nullptr;
public:
    enum WBS {
        non_call_wbs = 0x04,
        min_call_wbs = 0x08,
        interp_wbs = 0x0C,
        max_wbs      = 0x1C,
        default_wbs  = max_wbs
    }; // wbs for call

    enum BundleState {
        Open,
        Close,
        Unsafe,
    };

    enum WhichOperand {
        // input to locate_operand, and format code for relocations
        imm_operand  = 0,            // embedded 32-bit|64-bit immediate operand
        disp32_operand = 1,          // embedded 32-bit displacement or address
        call32_operand = 2,          // embedded 32-bit self-relative displacement
#ifndef _LP64
        _WhichOperand_limit = 3
#else
        narrow_oop_operand = 3,     // embedded 32-bit immediate narrow oop
        _WhichOperand_limit = 4
#endif
    };

    struct RegisterDelay {
        address header;
        int ticks;
    };

    // Maximimu size of an instruction for fast buffer code overflow checking.
    static const int max_bundle_size = 192;

protected:
    // The buffer into which code is generated. It could
    // either be owned by the assembler or be provided externally.
    byte* buffer_;
    unsigned int buffer_size_;
    bool is_external_buffer;

    // The program counter, which points into the buffer above and moves forward.
    byte* pc_;

    virtual void on_relocation(Relocation reloc, int offset, uint64_t value) = 0;

public:
    Assembler();
    Assembler(byte* buffer, unsigned offset, unsigned buffer_size);

    virtual ~Assembler();

    void set_buffer(byte* buffer, unsigned offset, unsigned buffer_size);
    byte* pc() const { return pc_; }
    int32_t pc_offset() const { return static_cast<int32_t>(pc_ - buffer_); }
    byte* buffer() const { return buffer_; }

   void init_asm_sheduler(int shed_level, bool sched_debug, bool patch_offsets, int free_reg, int free_registers_cnt, int insts_cnt);
   void finish_asm_sheduler();
   void e2k_print_asm_info();
   void e2k_asm_patch_rbs(int32_t *addr, int32_t new_rbs);
   void e2k_asm_patch_wsz(int32_t *addr, int32_t new_wsz);
   void def(E2kOpCode op, uint16_t exec_unit, uint32_t reg_code);
   int32_t get_lat(E2kOpCode op, uint16_t exec_unit, uint32_t reg_code);
   int32_t get_bypass_lat(E2kOpCode op, bool bypass, uint32_t reg_code);
   int32_t get_distance(uint32_t reg_code);
   void set_current_latency(int val);
   bool check_lts(int64_t val, DataTypes data_type, bool fixed_size, int reloc_kind);
   int32_t fits_in_size(int64_t val, DataTypes data_type, bool fixed_size, int reloc_kind);
   inline bool isInt(DataTypes data_type) { return data_type <= DATA_U64; };
   inline bool isSigned(DataTypes data_type) {return data_type %2 != 0; };

    static inline uint64_t elbrus_swap64(uint64_t val) {
        return ((val & 0xFFFFFFFF00000000ul) >> 32) | ((val & 0x00000000FFFFFFFFul) << 32);
    }
    static Condition invert_condition(Condition);
    static void flush_cache(void *start, size_t size_t);

    void emitb(uint8_t x);
    void emitw(uint16_t x);
    void emitl(uint32_t x);
    void emitq(uint64_t x, Relocation kind);

    void start();
    void close();
    void close_bundle();
    void start_bundle();
    void sched_close();

    uint32_t get_cond_type(int64_t pr_idx, bool inverse);
    uint32_t get_pr(int64_t pr);

    //operands encoding
    uint32_t SRC1(bool opnd_reg, int64_t opnd_val);
    uint32_t SRC2(bool opnd_reg, int64_t opnd_val, DataTypes data_type, bool fixed_size, int reloc_kind, int32_t als_id);
    uint32_t SRC3(bool opnd_reg, int64_t opnd_val);
    uint32_t DST(bool opnd_reg, int64_t opnd_val);
    uint32_t DST2(uint32_t cmpopce, bool opnd_reg, int64_t opnd_val);
    uint32_t ELP(bool opnd_reg, int64_t opnd_val);
    uint32_t get_cond_type(const Register& pr, bool inverse);
    uint32_t get_pr(const Register& pr);

    uint32_t LTS0_SPECIAL(uint32_t type, uint32_t rpsz, uint32_t wsz, uint32_t nfx, uint32_t dbl);
    bool new_lts(LtsType type, uint64_t val, int32_t als_idx = -1, Relocation kind = RelocationNone);
    void prepare_lts();
    void encode_lts();

    void prepare_cd();
    void prepare_len();
    void prepare_hs();

    //syllables encoding
    void ALF6(int32_t als_id, uint32_t opc, uint32_t fld1, uint32_t fld3, bool speculative = false);
    void ALF1_8(int32_t als_id, uint32_t opc, uint32_t fld1, uint32_t fld2, uint32_t fld3, bool speculative = false);
    void ALEF1_2(int32_t als_id, uint32_t opc2, uint32_t fld1);
    void C0F2(uint32_t ctpr, uint32_t ctp_opc, uint32_t disp);
    void C1F1(uint32_t opc, uint32_t param);
    void PL0F1(uint32_t elp0, uint32_t elp1, uint32_t clp0);
    void PL1F1(uint32_t elp2, uint32_t elp3, uint32_t clp1);
    void SF1(uint32_t ipd, uint32_t eap,  uint32_t bap, uint32_t srp, uint32_t vfdi, uint32_t crp, uint32_t abg,  uint32_t abn, uint32_t abp, uint32_t alc, uint32_t aa,  uint32_t ctop, uint32_t ctcond);

    void PLOPF0(uint32_t opc, uint32_t neg0, int64_t pr0_idx, uint32_t neg1,
         int64_t pr1_idx, uint32_t vdst, int64_t dst_idx);
    void PLOPF1(uint32_t opc, uint32_t neg0, int64_t pr0_idx, uint32_t neg1,
         int64_t pr1_idx, uint32_t vdst, int64_t dst_idx);
    void ALOPF1(E2kOpCode opc, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc, 
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF1(uint32_t opc, uint32_t als_id, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void CDOPF1(uint32_t opc, uint32_t als_id, int64_t pr_idx, bool inverse = false);
    void ALOPF_LD(E2kOpCode opc, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc, 
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF_LD_MAS(E2kOpCode opc, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc, 
         int64_t def_idx, uint32_t mas, bool inverse = false, bool speculative = false);
    void ALOPF2(uint32_t opc, uint32_t opce, uint32_t als_id, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF2(E2kOpCode opc, uint32_t opce, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF2ext(uint32_t opc, uint32_t opce, uint32_t opc2, uint32_t opce2,
         uint32_t als_id, int64_t pr_idx, bool use_reg, int64_t use_idx,
         DataTypes use_type, bool fixed_size, int use_reloc, int64_t def_idx, bool inverse = false);
    void ALOPF2ext(E2kOpCode opc, uint32_t opce, uint32_t opc2, uint32_t opce2,
         int64_t pr_idx, bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false);
    void ALOPF3(uint32_t opc, uint32_t als_id, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use_reloc,
         bool use3_reg, int64_t use3_idx, bool inverse = false, bool speculative = false);
    void ALOPF3(E2kOpCode opc, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx, 
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use_reloc, 
         bool use3_reg, int64_t use3, bool inverse = false, bool speculative = false);
    void ALOPF3_MAS(E2kOpCode opc, uint32_t als_id, int64_t pr_idx, bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use_reloc,
         bool use3_reg, int64_t use3_idx, 
         uint32_t mas,  bool inverse , bool speculative );
    void ALOPF3_MAS(E2kOpCode opc, uint32_t als_id, int64_t pr_idx, bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         bool use3_reg, int64_t use3_idx, 
         uint32_t mas,  unsigned disp, bool inverse, bool speculative);
    void ALOPF7(uint32_t opc, uint32_t cmpopce, uint32_t als_id, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF8(uint32_t opc, uint32_t cmpopce, uint32_t als_id, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF7(E2kOpCode opc, uint32_t cmpopce, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF8(E2kOpCode opc, uint32_t cmpopce, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF11(uint32_t opc, uint32_t opce, uint32_t opx, uint32_t als_id, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF11(E2kOpCode opc, uint32_t opce, uint32_t opx, int64_t pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF12(uint32_t opc, uint32_t opce, uint32_t als_id, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF12(E2kOpCode opc, uint32_t opce, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF15(uint32_t opc, uint32_t opce, uint32_t als_id, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         uint32_t reg_code, bool inverse = false, bool speculative = false);
    void ALOPF15(E2kOpCode opc, uint32_t opce, int64_t pr_idx,
         bool use_reg, int64_t use_idx, DataTypes use_type, bool fixed_size, int use_reloc,
         uint32_t reg_code, bool inverse = false, bool speculative = false);
    void ALOPF16(uint32_t opc, uint32_t opce, uint32_t als_id, int64_t pr_idx,
         uint32_t reg_code, int64_t def_idx, bool inverse = false, bool speculative = false);
    void ALOPF16(E2kOpCode opc, uint32_t opce, int64_t pr_idx,
         uint32_t reg_code, int64_t def_idx, bool inverse = false, bool speculative = false);
    void MERGE(E2kOpCode opc, int64_t pr_idx, int64_t m_pr_idx,
         bool use1_reg, int64_t use1_idx,
         bool use2_reg, int64_t use2_idx, DataTypes use2_type, bool fixed_size, int use2_reloc,
         int64_t def_idx, bool minverse, bool cinverse, bool speculative);
    void PLOPF(uint32_t opc, uint32_t neg0, int64_t pr0_idx, uint32_t neg1,
         int64_t pr1_idx, uint32_t vdst, int64_t dst_idx);
    void COPF2(int64_t ctpr_idx, uint32_t ctp_opc, unsigned disp);
    void COPF2(uint32_t opc, int64_t def_idx, unsigned disp);
    void COPF4(uint32_t opc, int64_t use_idx);
    void COPF4encode(uint32_t opc, int64_t param);
    void COPF4extEncode_new(uint32_t opc, int64_t param, int64_t type, int64_t rpsz,
         int64_t wsz, int64_t nfx, int64_t dbl);
    void COPF4ext(uint32_t opc, int64_t param_val, int64_t type_val,
         int64_t rpsz_val, int64_t wsz_val, int64_t nfx_val, unsigned dbl);
    void ct(int64_t pr_idx, int64_t ctpr_idx, bool inverse = false);
    void CT(int64_t pr_idx, int64_t ctpr_idx, bool inverse = false);
    void CALL(int64_t pr_idx, int64_t ctpr_idx, int64_t wbs_val, bool inverse = false);
    //operations encoding
    void ALOPF1(E2kOpCode opc, const Register pr, const Operand use1, const Operand use2, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF_LD(E2kOpCode opc, const Register pr, const Operand use1, const Operand use2, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF_LD_MAS(E2kOpCode opc, const Register pr, const Operand use1, const Operand use2, const Register def, uint32_t mas, bool inverse = false, bool speculative = false);
    void ALOPF2(E2kOpCode opc, uint32_t opce, const Register pr, const Operand use, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF2ext(E2kOpCode opc, uint32_t opce, uint32_t opc2, uint32_t opce2, const Register pr, const Operand use, const Register def, bool inverse = false);
    void ALOPF3(E2kOpCode opc, const Register pr, const Operand use1, const Operand use2, const Operand use3, bool inverse = false, bool speculative = false);
    void ALOPF_3_MAS(E2kOpCode opc, const Register pr, const Operand use1, const Operand use2, const Operand use3, uint32_t mas, Label* l, bool inverse = false, bool speculative = false);
    void ALOPF_3_MAS(E2kOpCode opc, const Register pr, const Operand use1, const Operand use2, const Operand use3, uint32_t mas, bool inverse = false, bool speculative = false);
    void ALOPF7(E2kOpCode opc, uint32_t cmpopce, const Register pr, const Operand use1, const Operand use2, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF8(E2kOpCode opc, uint32_t cmpopce, const Register pr, const Operand use, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF11(E2kOpCode opc, uint32_t opce, uint32_t opx, const Register pr, const Operand use1, const Operand use2, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF12(E2kOpCode opc, uint32_t opce, const Register pr, const Operand use, const Register def, bool inverse = false, bool speculative = false);
    void ALOPF15(E2kOpCode opc, uint32_t opce, const Register pr, const Operand use, uint32_t reg_code, bool inverse = false, bool speculative = false);
    void ALOPF16(E2kOpCode opc, uint32_t opce, const Register pr, uint32_t reg_code, const Register def, bool inverse = false, bool speculative = false);
    void MERGE(E2kOpCode opc, const Register pr, const Register m_pr, const Operand use1, const Operand use2, const Register def, bool minverse, bool cinverse, bool speculative);
    void COPF2(const Register ctpr, uint32_t ctp_opc, unsigned disp);
    void COPF2(uint32_t opc, const Register def, unsigned disp);
    void COPF4encode(uint32_t opc, unsigned param);
    void COPF4(uint32_t opc, unsigned use);
    void COPF4extEncode(uint32_t opc, unsigned param, unsigned type, unsigned rpsz, unsigned wsz, unsigned nfx, unsigned dbl);
    void COPF4ext(uint32_t opc, unsigned param, unsigned type, unsigned rpsz, unsigned wsz, unsigned nfx, unsigned dbl);
    void CALL(const Register pr, const Register ctpr, unsigned wbs, bool inverse = false);
    void ct(const Register pr, const Register ctpr, bool inverse = false);
    void CT(const Register pr, const Register ctpr, bool inverse = false);
    void PLOPF(uint32_t opc, uint32_t neg0, const Register pr0, uint32_t neg1, const Register pr1, uint32_t vdst, const Register dst);

    void set_als_available(uint16_t op_mask);
    bool try_als_available(uint16_t op_mask);
    bool valid_lts();
    int try_add_ST_ALU(uint32_t op);
    int try_add_LD_ALU(uint32_t op);
    int try_add_OP_ALU(uint32_t op);
    int try_add_OPext_ALU(uint32_t op, uint32_t opx);

    bool check_lts(const Operand src2);
    bool is_empty_bundle() const;
    void new_bundle();
    void guarantee_new_bundle();  // use to separate data-flow

    // Generic instructions
    // Does 32bit or 64bit as needed for the platform. In some sense these
    // belong in macro assembler but there is no need for both varieties to exist

    static int wbs_to_wsz(int wbs) {
        return ((wbs * 2) + 8) / 2;
    }

    static int wd_to_wsz(uint64_t wd) {
        return (wd >> 20) & 0xFF;
    }

    static int param_for_wsz(int n, int wsz) {
        return (wsz * 2 - 8) + n;
    }

    static int param_for_wbs(int n, int wbs) {
        return wbs * 2 + n;
    }

    static int make_getfs_const(unsigned from, unsigned size, bool sign) {
        from &= 0x1f;
        size &= 0x1f;
        unsigned sign_bit = sign ? 1U : 0U;
        return ((from << 0) | (size << 6) | (sign_bit << 12));
    }

    void bind(Label* L);  // binds an unbound label L to the current code position

    void nop(uint32_t nop = 2);
    void movd(const Operand src, const Register dst);
    void movs(const Operand src, const Register dst);
    void movd_pr(const Operand src, const Register dst, const Register pred, bool inv = false, bool spec = false);
    void movs_pr(const Operand src, const Register dst, const Register pred, bool inv = false, bool spec = false);

    AlsOffset mov_imm_encode_offset(const Immediate imm, const Register dst);
    uint32_t decode_als_get_lts_offset(AlsOffset encoded_offset);
    uint32_t decode_als_get_lts64_offset(AlsOffset encoded_offset);
    uint64_t patch_lts64(AlsOffset encoded_offset, uint64_t value);
    uint32_t patch_lts32(AlsOffset encoded_offset, uint32_t value);
#ifdef ASM_E2K_DEBUG
    inline void patch_lts32_dcheck(AlsOffset encoded_offset, uint32_t value, uint32_t expected);
    inline void patch_lts64_dcheck(AlsOffset encoded_offset, uint64_t value, uint64_t expected);
#else
#define patch_lts32_dcheck(encoded_offset, value, expected) patch_lts32(encoded_offset, value)
#define patch_lts64_dcheck(encoded_offset, value, expected) patch_lts64(encoded_offset, value)
#endif

    void push(const Operand src);
    void pop(const Register dst);

    // These are dummies to prevent surprise implicit conversions to Register
    void push(void* v);
    void pop(void* v);

    // Vanilla instructions in lexical order
    void disp(Label* l, const Register dst);
    void sdisp(uint32_t sdisp_num, const Register dst);
    void branch(const Register dst = EREG_CTPR1);
    void branch_cond(const Register pr, bool inverted, const Register dst = EREG_CTPR1);
    void return_ct(const Register pr = EREG_PRX, bool inverted = false);

    void ibranch_encode(const Register pr, unsigned disp, bool inverse = false);
    void ibranch_encode(int64_t pr_idx, unsigned disp, bool inverse = false);
    void ibranch(unsigned disp_in_words, const Register pr = EREG_PRX, bool inverse = false);
    void ibranch(Label* l, const Register pr = EREG_PRX, bool inverse = false);
    void call(const Register ctpr, unsigned wbs, const Register pr = EREG_PRX, bool inverse = false);

    void movtd(const Operand src, const Register dst);
    void lzcntd(const Operand src, const Register dst, const Register pred = EREG_PRX, bool inverse = false, bool spec = false);
    void lzcnts(const Operand src, const Register dst, const Register pred = EREG_PRX, bool inverse = false, bool spec = false);
    void popcntd(const Operand src, const Register dst, const Register pred = EREG_PRX, bool inverse = false, bool spec = false);
    void popcnts(const Operand src, const Register dst, const Register pred = EREG_PRX, bool inverse = false, bool spec = false);

    // generates appropriate comparison instruction, returns inverse flag
    bool cmp(Condition cc, const Operand src1, const Operand src2, const Register dst, bool double_word = true);
    // jcc is the generic conditional branch generator to run-
    // time routines, jcc is used for branches to labels. jcc
    // takes a branch opcode (cc) and a label (L) and generates
    // either a backward branch or a forward branch and links it
    // to the label fixup chain. Usage:
    //
    // Label L;      // unbound label
    // jcc(cc, L);   // forward branch to unbound label
    // bind(L);      // bind label to the current pc
    // jcc(cc, L);   // backward branch to bound label
    // bind(L);      // illegal: a label may be bound only once
    //
    // Note: The same Label can be used for forward and backward branches
    // but it may be bound only once.

    void jcc(Condition cc, Label* L, bool maybe_short = true);
    void jcc(Condition cc, const Operand src1, const Operand src2, Label* L, bool double_word = true);

    // Label operations & relative jumps (PPUM Appendix D)
    void jmp(Label* L);   // unconditional jump to L

    void jmp(const Operand entry); // pc <- entry
    void jmp(address entry);

    void lock();

    enum Membar_mask_bits {
        StoreStore = 1 << 3,
        LoadStore  = 1 << 2,
        StoreLoad  = 1 << 1,
        LoadLoad   = 1 << 0
    };

    // Writes a single word of data in the code stream.
    // Used for inline tables, e.g., jump-tables.
    void db(uint8_t data) { ASM_E2K_UNUSED(data); ASM_E2K_NOT_IMPLEMENTED(); } // TODO
    void dd(uint32_t data) { ASM_E2K_UNUSED(data); ASM_E2K_NOT_IMPLEMENTED(); } // TODO
    void dq(uint64_t data) { ASM_E2K_UNUSED(data); ASM_E2K_NOT_IMPLEMENTED(); } // TODO
    void dp(uintptr_t data) { dq(data); } // TODO
    void dq(Label* label) { ASM_E2K_UNUSED(label); ASM_E2K_NOT_IMPLEMENTED(); } // TODO

    // Serializes memory and blows flags
    void membar(Membar_mask_bits order_constraint) {
        // We only have to handle StoreLoad
        if (order_constraint & StoreLoad) {
            // All usable chips support "locked" instructions which suffice
            // as barriers, and are much faster than the alternative of
            // using cpuid instruction. We use here a locked add [esp],0.
            // This is conveniently otherwise a no-op except for blowing
            // flags.
            // Any change to this code may need to revisit other places in
            // the code where this idiom is used, in particular the
            // orderAccess code.
            lock();
            // addl(address(rsp, 0), 0);// Assert the lock# signal here
        }
    }

    void return_pp();
    void ret();
    void ret(const Operand value);
    void setwd(unsigned int wd);
    void setbn(unsigned rbs, unsigned rsz, unsigned rcur);

    void std(const Operand dst, const Operand disp, const Register value, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void stw(const Operand dst, const Operand disp, const Register value, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void sth(const Operand dst, const Operand disp, const Register value, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void stb(const Operand dst, const Operand disp, const Register value, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void std_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void stw_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void sth_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void stb_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void std_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, Label* l, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void stw_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, Label* l, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void sth_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, Label* l, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void stb_mas(const Operand dst, const Operand disp, const Register value, uint32_t mas, Label* l, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void ldd(const Operand src, const Operand disp, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ldw(const Operand src, const Operand disp, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ldh(const Operand src, const Operand disp, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ldb(const Operand src, const Operand disp, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void ldd_mas(const Operand src, const Operand disp, const Register dst, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ldw_mas(const Operand src, const Operand disp, const Register dst, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ldh_mas(const Operand src, const Operand disp, const Register dst, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ldb_mas(const Operand src, const Operand disp, const Register dst, uint32_t mas, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    // predicative arithmetic
    void addd_pr(const Operand src1, const Operand src2, const Register dst, const Register pr, bool inverse = false, bool speculative = false);
    void adds_pr(const Operand src1, const Operand src2, const Register dst, const Register pr, bool inverse = false, bool speculative = false);

    // arithmetic
    void addd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void addd_fd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void subd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void subd_fd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void muld(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void divd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void udivd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void mod(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void andd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void andnd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ord(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ornd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void xord(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void xornd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void shld(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void shrd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void scld(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void scrd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void sard(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void getfd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void merged(const Register mpred, const Operand src1, const Operand src2, const Register dst, bool minverse = false,
                const Register cpred = EREG_PRX, bool cinverse = false, bool spec = false);

    void adds(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void adds_fw(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void adds_fh(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void adds_fb(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void subs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void subs_fw(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void subs_fh(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void subs_fb(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void muls(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void smulh(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void umulh(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void smulx(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void umulx(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void smulx_fw(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void umulx_fw(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void umulx_fh(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void umulx_fb(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void smulx_fh(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void smulx_fb(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void divs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void udivs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void umodx(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void smodx(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ands(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void andns(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void ors(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void orns(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void xors(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void xorns(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void shls(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void shrs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void scls(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void scrs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void sars(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void getfs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void merges(const Register mpred, const Operand src1, const Operand src2, const Register dst, bool minverse = false,
                const Register cpred = EREG_PRX, bool cinvert = false, bool spec = false);

    // comparison
    void cmpd_e(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, e - equal
    void cmpandd_e(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, e - equal
    // signed comparison
    void cmpd_l(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, l - less
    void cmpd_le(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false); // d - double word, le - less or equal

    void cmpandd_le(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, le - less or equal
    // unsigned comparison
    void cmpd_b(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, b - below
    void cmpd_be(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false); // d - double word, le - below or equal

    void cmps_e(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);     // s - single word, e - equal
    void cmpands_e(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // s - single word, e - equal

    // signed comparison
    void cmps_l(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // s - single word, l - less
    void cmps_le(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false); // s - single word, le - less or equal

    void cmpands_le(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // s - single word, le - less or equal
    // unsigned comparison
    void cmps_b(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // s - single word, b - below
    void cmps_be(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false); // s - single word, le - below or equal

    void cmpands_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // s - single word, s - less then zero
    void cmpands_p(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // s - single word, p - odd

    void cmpandd_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, s - less then zero
    void cmpandd_p(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);  // d - double word, p - odd

    // checks for flag in IFL format
    void cctopo(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctopb(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctope(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctopbe(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctops(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctopp(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctopl(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void cctople(const Operand flag, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    // non predicative floating point arithmetic
    void faddd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fsubd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fmuld(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdivd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void fmaxs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fmins(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fmaxd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fmind(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void fadds(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fsubs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fmuls(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdivs(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    // floating point comparison
    void fcmpd_e(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_e(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_l(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_l(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_b(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_b(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_le(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_le(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_be(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_be(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_uod(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_uod(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_od(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_od(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    // floating point comparison
    void fcmpd_e_d(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_e_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_l_d(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_l_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_le_d(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_le_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_uod_d(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_uod_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_od_d(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_od_s(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void fcmpd_od_f(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_od_f(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmpd_ud_f(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fcmps_ud_f(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    // convertions
    void istofs(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void istofd(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void idtofs(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void idtofd(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void fstois(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fstoistr(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fstoid(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fstoidtr(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fstofd(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdtois(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdtoistr(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdtoid(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdtoidtr(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdtofs(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fdtoifd(const Operand src1, const Register src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void sxt(const Operand use1, const Operand use2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void fsqrts(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void fsqrtid(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void fsqrttd(const Operand src1, const Operand src2, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void gettagd(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void gettags(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void wait(unsigned value);
    void getsp(const Operand src, const Register dst, const Register pred = EREG_PRX, bool invert = false);

    void rrd(uint32_t reg, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void rrs(uint32_t reg, const Register dst, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void rwd(const Operand src, uint32_t reg, const Register pred = EREG_PRX, bool invert = false, bool spec = false);
    void rws(const Operand src, uint32_t reg, const Register pred = EREG_PRX, bool invert = false, bool spec = false);

    void andp(const Register src1, bool neg1, const Register src2, bool neg2, const Register dst);
    void landp(const Register src1, bool neg1, const Register src2, bool neg2, const Register dst);
    void movep(const Register pred, bool negp, const Register src, bool negs, const Register dst);
    static bool is_polling_page_far() { return true; }

    /**
   * @brief compute cs0 offset by header
   * @param hs_s
   * @param hs_al
   * @return cs0 offset
   */
    static int compute_cs0_offset_by_header(uint32_t hs_s, uint32_t hs_al);

    /**
   * @brief get offset for disp
   * @param address first
   * @param address second
   * @return offset in from for CS0.DISP
   */
    static int32_t get_offset_for_disp(address first, address second);

    static const int _c1f1_shift[];
protected:
    uint16_t  _als_available; // exec units available in this bundle
    uint16_t  _st_mask;
    uint16_t  _ld_mask;
    uint16_t  _mem_mask;
    uint16_t  _lts_size;
    uint16_t  _lts16_num;
#if ASM_E2K_LEGACY_GUARANTEE_DELAY
    RegisterDelay _pred_delays[PredicatesUsedCount];
    RegisterDelay _control_delays[ControlRegistersCount];
#endif
    int _tick;

    BundleState _bundle_state;

    PredDefMap defs;
    LatencyMap latency;
    int32_t prev_pc_offset;

    BitWriter bw;
protected:
    // code emission

    // Avoid using directly section
    // Instructions in this section are actually usable by anyone without danger
    // of failure but have performance issues that are addressed my enhanced
    // instructions which will do the proper thing base on the particular cpu.
    // We protect them because we don't trust you...

    void clear_defmaps();
    void update_latency();

    int to_als_id(uint16_t exec_unit);
    int get_exec_unit(uint16_t mask, bool is_load, bool is_store);
    int get_exec_u_ld(uint16_t mask);
    int get_exec_u_st(uint16_t mask);

#if ASM_E2K_LEGACY_GUARANTEE_DELAY
    void do_tick();
    void reset_delay(const Operand op);
    void guarantee_delay(const Operand op, int tick);
#endif

    void set_def(const uint32_t def_idx, const uint32_t pred_idx, bool inverse);
    void set_uncond_def(const uint32_t def_idx);
    void set_def(const Register& def, const Register& pred, bool inverse);
    void set_uncond_def(const Register& def);

    void bind_to(Label* L, int pos);
    bool check_uses_prx(const uint32_t use1_idx, bool use1_reg);
    bool check_uses(const uint32_t use1_idx, bool use1_reg,
         const uint32_t pred_idx, bool inverse);
    bool check_uses_prx(const uint32_t use1_idx, bool use1_reg,
         const uint32_t use2_idx, bool use2_reg); 
    bool check_uses(const uint32_t use1_idx, bool use1_reg,
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t pred_idx, bool inverse);
    bool check_uses_prx(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg);
    bool check_uses(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t pred_idx, bool inverse);
    bool check_uses_prx(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t use4_idx, bool use4_reg);
    bool check_uses(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t use4_idx, bool use4_reg,
         const uint32_t pred_idx, bool inverse);
    bool check_uses_prx(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t use4_idx, bool use4_reg,
         const uint32_t use5_idx, bool use5_reg);
    bool check_uses(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t use4_idx, bool use4_reg,
         const uint32_t use5_idx, bool use5_reg,
         const uint32_t pred_idx, bool inverse);
    bool check_uses_prx(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t use4_idx, bool use4_reg,
         const uint32_t use5_idx, bool use5_reg,
         const uint32_t use6_idx, bool use6_reg);
    bool check_uses(const uint32_t use1_idx, bool use1_reg, 
         const uint32_t use2_idx, bool use2_reg, 
         const uint32_t use3_idx, bool use3_reg,
         const uint32_t use4_idx, bool use4_reg,
         const uint32_t use5_idx, bool use5_reg,
         const uint32_t use6_idx, bool use6_reg,
         const uint32_t pred_idx, bool inverse);

    /**
   * @brief parse command by command_offset and patch CS0.DISP
   * @return offset of next instruction to patch
   *
   */
    int32_t patch_command_disp(int command_offset, int label_offset);

    virtual void grow_buffer();

    virtual inline bool buffer_overflow() const {
        return pc_ >= buffer_ + buffer_size_ - max_bundle_size;
    }

};

class E2kInstructionRegion {
public:
    Assembler* assembler;
    E2kInstructionRegion(Assembler* masm);
    ~E2kInstructionRegion();
private:
    address _begin;
    bool _in_emitter;
};
extern thread_local AsmScheduler* asm_sched;

typedef struct {
# define SCHED_E2K_ASSEMBLER_SIZE 256000
    static_assert(sizeof(AsmScheduler) <= SCHED_E2K_ASSEMBLER_SIZE, "Assembler class size is changed");
    uint8_t data[SCHED_E2K_ASSEMBLER_SIZE];
} sched_e2k_handle;
extern thread_local sched_e2k_handle sched_buf;
#endif // __cplusplus

#ifdef ASM_E2K_GENERATE_C_API

#ifdef __cplusplus
# define ASM_E2K_EXTERN_C extern "C"
#else
# define ASM_E2K_EXTERN_C extern
#endif

typedef uint64_t Opnd;
typedef struct {
# define ASM_E2K_ASSEMBLER_SIZE 5000
#ifdef __cplusplus
    static_assert(sizeof(Assembler) <= ASM_E2K_ASSEMBLER_SIZE, "Assembler class size is changed");
#endif
    uint8_t data[ASM_E2K_ASSEMBLER_SIZE];
} asm_e2k_handle;

//TODO: tmp
//template<int s> struct Wow;
//Wow<sizeof(Assembler)> wow;

typedef enum {
    OP_REG  = 0,
    OP_IMMI = DATA_I64,
    OP_IMMU = DATA_U64,
    OP_IMMF = DATA_S,
    OP_IMMD = DATA_D
} OpndKind;

#define Reg(r)  (r), OP_REG
#define ImmF(r) (float)(r), OP_IMMF
#define ImmD(r) (double)(r), OP_IMMD
#define ImmI(r) (uint64_t)(r), OP_IMMI
#define ImmU(r) (uint64_t)(r), OP_IMMU
#define ImmL(r) (uint64_t)(r), OP_IMMI
#define ImmUL(r) (uint64_t)(r), OP_IMMU
#define OpndArg(op) Opnd op, OpndKind op## _t
#define OpndUse(op) op, op## _t

ASM_E2K_EXTERN_C asm_e2k_handle* asm_e2k_init(asm_e2k_handle* h, uint8_t* buffer, unsigned pc_offset, unsigned buffer_size);
ASM_E2K_EXTERN_C asm_e2k_handle* e2k_shed_cg_init(int shed_level, bool sched_debug, bool patch_offsets, int insts_cnt,
                                 int free_reg, int free_registers_cnt, asm_e2k_handle* h,
                                 uint8_t* buffer,unsigned offset, unsigned buffer_size);
ASM_E2K_EXTERN_C void e2k_init_asm_statistic(asm_e2k_handle* h);
ASM_E2K_EXTERN_C int e2k_get_rbs_offsets(asm_e2k_handle* h, long *arr_rbs_offsets, int *free_regs_cnt);

ASM_E2K_EXTERN_C int  e2k_finish_asm_statistic(asm_e2k_handle* h, int *nop_cnt);
ASM_E2K_EXTERN_C void e2k_asm_patch_rbs(asm_e2k_handle* h, int32_t *addr, int32_t new_rbs);
ASM_E2K_EXTERN_C void e2k_asm_patch_wsz(asm_e2k_handle* h, int32_t *addr, int32_t new_wsz);
ASM_E2K_EXTERN_C int e2k_get_wsz_offset_after_setwd(int32_t addr);
ASM_E2K_EXTERN_C int e2k_get_rbs_offset_after_setwd(int32_t addr);
ASM_E2K_EXTERN_C void asm_e2k_set_buf(asm_e2k_handle* h, uint8_t* buffer, unsigned pc_offset, unsigned buffer_size);
ASM_E2K_EXTERN_C void asm_e2k_start(asm_e2k_handle* h);
ASM_E2K_EXTERN_C void asm_e2k_close(asm_e2k_handle* h);

ASM_E2K_EXTERN_C void Label_init(Label* l);
ASM_E2K_EXTERN_C void Label_bind(Label* l, asm_e2k_handle* h);
ASM_E2K_EXTERN_C void Label_bind_loc(Label* l, int offset);

ASM_E2K_EXTERN_C void start_bundle(asm_e2k_handle* h);
ASM_E2K_EXTERN_C void close_bundle(asm_e2k_handle* h);
ASM_E2K_EXTERN_C void guarantee_new_bundle(asm_e2k_handle* h);
ASM_E2K_EXTERN_C int  is_empty_bundle(asm_e2k_handle* h);

ASM_E2K_EXTERN_C int pc_offset(asm_e2k_handle* h);
ASM_E2K_EXTERN_C unsigned char* pc(asm_e2k_handle* h);
ASM_E2K_EXTERN_C unsigned char* asm_e2k_buffer(asm_e2k_handle* h);

ASM_E2K_EXTERN_C int compute_cs0_offset_by_header(uint32_t hs_s, uint32_t hs_al);

ASM_E2K_EXTERN_C void nop(asm_e2k_handle* h, uint32_t nops);
ASM_E2K_EXTERN_C void call(asm_e2k_handle* h, Opnd pr, Opnd ctpr, Opnd wbs, bool inverse);
ASM_E2K_EXTERN_C void disp(asm_e2k_handle* h, Label *l, Opnd dst);
ASM_E2K_EXTERN_C void _disp(asm_e2k_handle* h, int32_t offset, Opnd dst);
ASM_E2K_EXTERN_C void branch(asm_e2k_handle* h, Opnd dst);
ASM_E2K_EXTERN_C void branch_cond(asm_e2k_handle* h, Opnd pr, bool inverted, Opnd dst);
ASM_E2K_EXTERN_C void return_ct(asm_e2k_handle* h, Opnd pr, bool inverted);
ASM_E2K_EXTERN_C void ibranch(asm_e2k_handle* h, Label* l, Opnd pr, bool inverse);
ASM_E2K_EXTERN_C void cmpd_e(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpandd_e(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpd_l(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpd_le(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpandd_le(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpd_b(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpd_be(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmps_e(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpands_e(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmps_l(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmps_le(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpands_le(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmps_b(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmps_be(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpands_s(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpands_p(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpandd_s(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cmpandd_p(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmpd_e(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmps_e(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmpd_l(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmps_l(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmpd_be(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmps_be(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmpd_le(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmps_le(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmpd_b(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmps_b(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fcmpd_uod(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void istofd(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void istofs(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fstois(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fdtois(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fdtoistr(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fdtoidtr(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fdivs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fdivd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void faddd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fadds(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fsubd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fsubs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fmuld(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fmuls(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fdtofs(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void fstofd(asm_e2k_handle* h, OpndArg(src1), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void sxt(asm_e2k_handle* h, OpndArg(use1), OpndArg(use2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void getsp(asm_e2k_handle* h, OpndArg(src), Opnd dst, Opnd pred, bool invert);
ASM_E2K_EXTERN_C void rrd(asm_e2k_handle* h, uint32_t reg, Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void rrs(asm_e2k_handle* h, uint32_t reg, Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void rwd(asm_e2k_handle* h, OpndArg(src), uint32_t reg, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void rws(asm_e2k_handle* h, OpndArg(src), uint32_t reg, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void movtd(asm_e2k_handle* h, OpndArg(src), Opnd dst);
ASM_E2K_EXTERN_C void jmp_l(asm_e2k_handle* h, Label *L);
ASM_E2K_EXTERN_C void jmp(asm_e2k_handle* h, OpndArg(entry));
ASM_E2K_EXTERN_C void addd_pr(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pr, bool inverse, bool speculative);
ASM_E2K_EXTERN_C void adds_pr(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pr, bool inverse, bool speculative);
ASM_E2K_EXTERN_C void addd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void adds(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void adds_fw(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void adds_fh(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void adds_fb(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void addd_fd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void mov(asm_e2k_handle* h, OpndArg(src), Opnd dst);
ASM_E2K_EXTERN_C void movd_pr(asm_e2k_handle* h, OpndArg(src), Opnd dst, Opnd pred, bool inv, bool spec);
ASM_E2K_EXTERN_C void movs_pr(asm_e2k_handle* h, OpndArg(src), Opnd dst, Opnd pred, bool inv, bool spec);
ASM_E2K_EXTERN_C void movs(asm_e2k_handle* h, OpndArg(src), Opnd dst);
ASM_E2K_EXTERN_C void return_pp(asm_e2k_handle* h);
ASM_E2K_EXTERN_C void ret_void(asm_e2k_handle* h);
ASM_E2K_EXTERN_C void ret(asm_e2k_handle* h, OpndArg(value));
ASM_E2K_EXTERN_C void std(asm_e2k_handle* h, OpndArg(dst), OpndArg(disp), Opnd src, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void stw(asm_e2k_handle* h, OpndArg(dst), OpndArg(disp), Opnd src, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void sth(asm_e2k_handle* h, OpndArg(dst), OpndArg(disp), Opnd src, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void stb(asm_e2k_handle* h, OpndArg(dst), OpndArg(disp), Opnd src, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldd(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldw(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldh(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldb(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldd_mas(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, uint32_t mas, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldw_mas(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, uint32_t mas, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldh_mas(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, uint32_t mas, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ldb_mas(asm_e2k_handle* h, OpndArg(src), OpndArg(disp), Opnd dst, uint32_t mas, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void subd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void subd_fd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void subs_fw(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void subs_fb(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void muld(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void divd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void udivd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void mod(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void andd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void andp(asm_e2k_handle* h, Opnd src1, bool neg1, Opnd src2, bool neg2, Opnd dst);
ASM_E2K_EXTERN_C void landp(asm_e2k_handle* h, Opnd src1, bool neg1, Opnd src2, bool neg2, Opnd dst);
ASM_E2K_EXTERN_C void movep(asm_e2k_handle* h, Opnd pred, bool negp, Opnd src, bool negs, Opnd dst);
ASM_E2K_EXTERN_C void ord(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ornd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void xord(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void shld(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void shrd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void scld(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void scrd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void sard(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void getfd(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void merged(asm_e2k_handle* h, Opnd mpred, OpndArg(src1), OpndArg(src2), Opnd dst, bool minverse, Opnd cpred, bool cinverse, bool spec);
ASM_E2K_EXTERN_C void subs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void muls(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void smulx(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void smulh(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void smulx_fw(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void smulx_fh(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void smulx_fb(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void umulx(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void umulh(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void umulx_fw(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void umulx_fh(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void umulx_fb(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void divs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void udivs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void umodx(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ands(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void ors(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void orns(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void xors(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void shls(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void shrs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void scls(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void scrs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void sars(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void getfs(asm_e2k_handle* h, OpndArg(src1), OpndArg(src2), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void merges(asm_e2k_handle* h, Opnd mpred, OpndArg(src1), OpndArg(src2), Opnd dst, bool minverse, Opnd cpred, bool cinverse, bool spec);
ASM_E2K_EXTERN_C void cctopo(asm_e2k_handle* h, OpndArg(flag), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void cctopb(asm_e2k_handle* h, OpndArg(flag), Opnd dst, Opnd pred, bool invert, bool spec);
ASM_E2K_EXTERN_C void setwd(asm_e2k_handle* h, unsigned wd);
ASM_E2K_EXTERN_C void setbn(asm_e2k_handle* h, unsigned rbs, unsigned  rsz, unsigned rcur);
ASM_E2K_EXTERN_C void wait_inst(asm_e2k_handle* h, Opnd use);
ASM_E2K_EXTERN_C void set_register_window(asm_e2k_handle* h, unsigned wsz, int need_bn);
#endif // ASM_E2K_GENERATE_C_API

NAMESPACE_ASM_E2K_END

#endif // ASM_E2K_ASSEMBLER_H
