#ifndef TANGO_ATOMIC_HPP
#define TANGO_ATOMIC_HPP

#include <cstdint>

namespace tango {
namespace atomic {

#define ALWAYS_INLINE __attribute__((always_inline))

inline void atomic_copy64(volatile void* src, volatile void* dst) ALWAYS_INLINE;
inline void store(int32_t store_value, volatile int32_t* dest) ALWAYS_INLINE;
inline void store(int64_t store_value, int64_t* dest) ALWAYS_INLINE;
inline void store(int64_t store_value, volatile int64_t* dest) ALWAYS_INLINE;
inline void store_ptr(intptr_t store_value, intptr_t* dest) ALWAYS_INLINE;
inline void store_ptr(intptr_t store_value, volatile intptr_t* dest) ALWAYS_INLINE;
inline int32_t add(int32_t add_value, volatile int32_t* dest) ALWAYS_INLINE;
inline intptr_t add_ptr(intptr_t add_value, volatile intptr_t* dest) ALWAYS_INLINE;
inline void* add_ptr(intptr_t add_value, volatile void* dest) ALWAYS_INLINE;
inline void inc(volatile int32_t* dest) ALWAYS_INLINE;
inline void inc_ptr(volatile intptr_t* dest) ALWAYS_INLINE;
inline void inc_ptr(volatile void* dest) ALWAYS_INLINE;
inline void dec(volatile int32_t* dest) ALWAYS_INLINE;
inline void dec_ptr(volatile intptr_t* dest) ALWAYS_INLINE;
inline void dec_ptr(volatile void* dest) ALWAYS_INLINE;
inline int32_t xchg(int32_t xchg_value, volatile int32_t* dest) ALWAYS_INLINE;
inline intptr_t xchg_ptr(intptr_t xchg_value, volatile intptr_t* dest) ALWAYS_INLINE;
inline void* xchg_ptr(void* xchg_value, volatile void* dest) ALWAYS_INLINE;
inline int32_t cmpxchg(int32_t xchg_value, volatile int32_t* dest, int32_t cmp_value) ALWAYS_INLINE;
inline int64_t cmpxchg(int64_t xchg_value, volatile int64_t* dest, int64_t cmp_value) ALWAYS_INLINE;
inline intptr_t cmpxchg_ptr(intptr_t xchg_value, volatile intptr_t* dest, intptr_t cmp_value) ALWAYS_INLINE;
inline void* cmpxchg_ptr(void* xchg_value, volatile void* dest, void* cmp_value) ALWAYS_INLINE;
inline int64_t load(volatile int64_t* src) ALWAYS_INLINE;

#undef ALWAYS_INLINE

//-------------------------------------------------------------------------------------//

inline void atomic_copy64(volatile void* src, volatile void* dst) {
  *(int64_t *) dst = *(int64_t *) src;
}

//-------------------------------------------------------------------------------------//

inline void store(int32_t store_value, volatile int32_t* dest) {
  *dest = store_value;
}

//-------------------------------------------------------------------------------------//

inline void store(int64_t store_value, int64_t* dest) {
  *dest = store_value;
}

//-------------------------------------------------------------------------------------//

inline void store(int64_t store_value, volatile int64_t* dest) {
  *dest = store_value;
}

//-------------------------------------------------------------------------------------//

inline void store_ptr(intptr_t store_value, intptr_t* dest) {
  *dest = store_value;
}

//-------------------------------------------------------------------------------------//

inline void store_ptr(intptr_t store_value, volatile intptr_t* dest) {
  *dest = store_value;
}

//-------------------------------------------------------------------------------------//

inline int32_t add(int32_t add_value, volatile int32_t* dest) {
  return __sync_add_and_fetch(dest, add_value);
}

//-------------------------------------------------------------------------------------//

inline intptr_t add_ptr(intptr_t add_value, volatile intptr_t* dest) {
  return __sync_add_and_fetch(dest, add_value);
}

//-------------------------------------------------------------------------------------//

inline void* add_ptr(intptr_t add_value, volatile void* dest) {
  return (void *) add_ptr(add_value, (volatile intptr_t *) dest);
}

//-------------------------------------------------------------------------------------//

inline void inc(volatile int32_t* dest) { add(1, dest); }
inline void inc_ptr(volatile intptr_t* dest) { add_ptr(1, dest); }
inline void inc_ptr(volatile void* dest) { add_ptr(1, dest); }
inline void dec(volatile int32_t* dest) { add(-1, dest); }
inline void dec_ptr(volatile intptr_t* dest) { add_ptr(-1, dest); }
inline void dec_ptr(volatile void* dest) { add_ptr(-1, dest); }

//-------------------------------------------------------------------------------------//

inline int32_t xchg(int32_t xchg_value, volatile int32_t* dest) {
  // __sync_lock_test_and_set is a bizarrely named atomic exchange
  // operation.  Note that some platforms only support this with the
  // limitation that the only valid value to store is the immediate
  // constant 1.  There is a test for this in JNI_CreateJavaVM().
  return __sync_lock_test_and_set (dest, xchg_value);
}

//-------------------------------------------------------------------------------------//

inline intptr_t xchg_ptr(intptr_t xchg_value, volatile intptr_t* dest) {
  return __sync_lock_test_and_set (dest, xchg_value);
}

//-------------------------------------------------------------------------------------//

inline void* xchg_ptr(void* xchg_value, volatile void* dest) {
  return (void *) xchg_ptr((intptr_t) xchg_value, (volatile intptr_t*) dest);
}

//-------------------------------------------------------------------------------------//

inline int32_t cmpxchg(int32_t xchg_value, volatile int32_t* dest, int32_t cmp_value) {
  return __sync_val_compare_and_swap(dest, cmp_value, xchg_value);
}

//-------------------------------------------------------------------------------------//

inline int64_t cmpxchg(int64_t xchg_value, volatile int64_t* dest, int64_t cmp_value) {
  return __sync_val_compare_and_swap(dest, cmp_value, xchg_value);
}

//-------------------------------------------------------------------------------------//

inline intptr_t cmpxchg_ptr(intptr_t xchg_value, volatile intptr_t* dest, intptr_t cmp_value) {
  return __sync_val_compare_and_swap(dest, cmp_value, xchg_value);
}

//-------------------------------------------------------------------------------------//

inline void* cmpxchg_ptr(void* xchg_value, volatile void* dest, void* cmp_value) {
  return (void *) cmpxchg_ptr((intptr_t) xchg_value, 
                              (volatile intptr_t*) dest,
                              (intptr_t) cmp_value);
}

//-------------------------------------------------------------------------------------//

inline int64_t load(volatile int64_t* src) {
  volatile int64_t dest;
  atomic_copy64(src, &dest);
  return dest;
}

//-------------------------------------------------------------------------------------//

}      // namespace atomic
}      // namespace tango
#endif // TANGO ATOMIC HPP
