#ifndef TANGO_GLOBAL_LOGGER_HPP
#define TANGO_GLOBAL_LOGGER_HPP

#include "tango/ir.hpp"
#include "tango/tango_inst.hpp"
#include "tango/tango_globals.hpp"

namespace tango {

//-------------------------------------------------------------------------------------//

//Need to identify which phase is logging & which phases are passed for verifier
enum PhaseName : uint32_t {
  NONE = 0,        // No phases
  
  BCL  = 1,   // Block Cloner
  CLA  = 2,   // Code Layouter
  CLS  = 3,   // Copy Coalescer
  CPN  = 4,   // Copy Propogation
  CRR  = 5,   // CondReadRegs
  DCE  = 6,   // Dead Code Elemination
  DCG  = 7,   // Dynamic Checks Generator
  EMR  = 8,   // Emitter
  FCC  = 9,   // Fast Copy Coalescer
  IFC  = 10,  // If Conversion
  ISR  = 11,  // Instruction Scheduler
  LAN  = 12,  // Live Analyser
  LBL  = 13,  // Loop Building
  LPP  = 14,  // Land Peep
  LTR  = 15,  // Loop Transformations
  MLI  = 16,  // Move Invariants Outside Loop
  MRG  = 17,  // Block Merge
  OMR  = 18,  // OOP Resolver
  PHL  = 19,  // Peephole
  PGR  = 20,  // Prolog Generator
  RAR  = 21,  // Register Allocator
  SDCE = 22,  // Ssa DeadCodeEliminator
  SEL  = 23,  // Selector
  SFR  = 24,  // State Flusher
  SGN  = 25,  // Spill Generator
  SPL  = 26,  // Software Pipeliner
  SSA  = 27,  // DeSSA
  SXT  = 28,  // SXT Corrector
  ATB  = 29,  // Tree Balancer
  CPM  = 30,  // Cold Path Merging
  MIS  = 31,  // Macro Instructions
  ALL  = 32,

  END  = ALL + 1   // need to define all without disjunction
};


//-------------------------------------------------------------------------------------//
// need to set level of logging
enum class LogLevel : uint8_t {
  _ASM,    // Only assembler, before and after phase
  _DEBUG,  // Only debug, which is set by developer in method body (macro LOG_DEBUG)
  _GRAPH,  // Only cfg printed to .dot file, before and after phase
  _FULL    // All 3 options
};

//-------------------------------------------------------------------------------------//

#define ENDL "\n"

//-------------------------------------------------------------------------------------//

struct LogOption {
  bool log_asm;
  bool log_debug;
  bool log_graph;
};

typedef arena_map<PhaseName, LogOption, std::less<PhaseName>, 
                  persistent_allocator>::type Options; //logging options

#ifdef TANGO_LOG_ENABLED

//-------------------------------------------------------------------------------------//

class GlobalLogger : public PersistentObj {
 private :
  typedef arena_vector<char, persistent_allocator>::type DelimiterVector;
  typedef arena_map<persistent_string, PhaseName, 
                    std::less<persistent_string>, persistent_allocator>::type   MapName;
  typedef arena_map<persistent_string, LogLevel, 
                    std::less<persistent_string>, persistent_allocator>::type   MapLevel;
  typedef arena_map<PhaseName, persistent_string, 
                    std::less<PhaseName>, persistent_allocator>::type     MapFullname;

  struct LogPair {
    PhaseName name;
    LogLevel  level;
  };
  // parse string TangoLog, and fill map Options, also need to resolve fullname of phase
  class OptionParser {
   public:
    OptionParser();
    LogPair                 getOption(arena_string&)                                const;

   private:
    arena_string            getOptionString(arena_string&)                          const;
    arena_string::size_type getDelimiterPos(arena_string&, const DelimiterVector&)  const;
    LogPair                 splitOptionString(arena_string&)                        const;

    void                    initPhaseName();
    void                    initLogLevel();
    void                    initPhaseFullname();

   public:
    MapFullname             phase_fullname;
    
   private:
    DelimiterVector         _string_delimiters;
    DelimiterVector         _option_delimiters;

    MapName                 _phase_name;
    MapLevel                _log_level;
  };

 public:
  static GlobalLogger& getInstance();

  void           initOptions(Options&, const char*);

  Options&       getOptions()                      { return _options;  }
  const char*    getFullname(PhaseName name) const { return _parser.phase_fullname.at(name).c_str(); }

 private:
  void           printMap(Options&) const;

 private:
  GlobalLogger();
  GlobalLogger(GlobalLogger const&) = delete;
  GlobalLogger(GlobalLogger&&) = delete;
  GlobalLogger& operator=(GlobalLogger const&) = delete;
  GlobalLogger& operator=(GlobalLogger&&) = delete;

 private:
  Options                 _options;
  OptionParser            _parser;
};

//-------------------------------------------------------------------------------------//

#endif  // TANGO LOG ENABLED
}       // namespace tango
#endif  // TANGO_GLOBAL_LOGGER_HPP
