#ifndef TANGO_INST_DEPENDENCY_HPP
#define TANGO_INST_DEPENDENCY_HPP

#include "tango/ir.hpp"
#include "tango/clone_helper.hpp"
#include "tango/local_logger.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class   DepEdge;
class   DepNode;

typedef arena_list     <DepNode*>::type                           DepNodeList;

//----------------------------------------------------------------------------//

class OpndDepNode {
public:
  DepNodeList uses;
  DepNodeList cond_reads;
  DepNode*    def;

  OpndDepNode(): def(NULL) {}
};


//----------------------------------------------------------------------------//

struct RegOpndComparator {
  bool operator()(const RegOpnd* a, const RegOpnd* b) const {
    if (a->location != UNALLOCATED || b->location != UNALLOCATED) {
      return a->location < b->location;
    }
    return a->id < b->id;
  }
};

//----------------------------------------------------------------------------//

typedef arena_list     <DepEdge*>::type                           DepEdgeList;
typedef arena_multimap <int32_t, DepNode*>::type                  Int2DepNodeMap;
typedef arena_map      <Inst*, DepNode*>::type                    Inst2DepNodeMap;
typedef arena_map<Inst*, int>::type                               InstToIntMap;
#if (__LCC__ > 122)
typedef arena_unordered_map<RegOpnd*, InstList>::type             RegToInstList;
#else
typedef arena_map<RegOpnd*, InstList>::type                       RegToInstList;
#endif
typedef arena_map<RegOpnd*, OpndDepNode, RegOpndComparator>::type RegOpndToDepNodeMap;

//----------------------------------------------------------------------------//

class LockCheckDisambiguator;
class RegisterAllocHelper;

//----------------------------------------------------------------------------//

enum DepType {
  DepTypeUnknown,
  DepTypeFlow,
  DepTypeAnti,
  DepTypeOutput,
  DepTypeRoot
};

//----------------------------------------------------------------------------//
#define OF_UNROLL_ALL_ITER        0
#define OF_UNROLL_FIRST_ITER      1
#define OF_CHECK_BOUND            1
#define OF_UNROLL_LAST_ITER       2
#define OF_LOCALIZED_MEM          2
#define OF_UNROLL_DUPLICATE_MEM   4
#define OF_ROTATE_BASE            4
#define OF_UNROLL_LIVE_AFTER_LOOP 8
#define OF_FOR_ADDRESS_EVALUATION 8
#define OF_UNROLL_USE_ONE_REGISTER 16
#define OF_INDUCTION               32
#define OF_REDUCTION               64
#define OF_USED_IN_UNLINEAR_EXP    128

#define SXT_NONE    0
#define SXT_SXT     1
#define SXT_ZXT     2
#define SXT_RAW_SXT 3
#define SXT_UNKNOWN 4

//class describes additional info which can be propagated during flow analysis
class PropResource : public TangoObj {
public:
    PropResource() : base_address(NULL), rotating_base(NULL), base_addr_inst_def(NULL),
        invariant_offset(NULL), offset(0), factor(0), dep_type(0), opt_flag(0), sxt_type(SXT_NONE) {}
    bool is_opt_flag(int val) const { return (opt_flag & val) != 0; }
    void set_opt_flag(int val) { opt_flag |= val; }
    void clear_opt_flag(int val) { opt_flag &= ~val; }
public:
    RegOpnd*    base_address; // for cannon expr: BASE_ADDR + FACTOR*iv + OFFSET
    RegOpnd*    rotating_base; // for cannon expr:rotating_base = FACTOR*iv
    RegOpnd*    invariant_offset; // invariant part of offset
    Inst*       base_addr_inst_def; // instruction where base address is defined
    int64_t     offset;   // for cannon expr: base_addr + factor*iv + OFFSET
    int32_t     factor;   // for cannon expr: base_addr + FACTOR*iv + OFFSET
    int32_t     opt_flag;
    int16_t     dep_type;
    int16_t     sxt_type; // no sxt = 0, sxt = 1, zxt = 2, raw_sxt = 3, not one sxt = 4
};

//----------------------------------------------------------------------------//

class DepNode : public TangoObj {
public:
  DepNode(Inst* inst, Block* b) : inst(inst), block(b), visited(false), weight(0), trg_cycle(0),
      dep_cnt(0), resource(NULL) {}
  void        setCycle(int32_t cycle) { if (trg_cycle < cycle) trg_cycle = cycle; }
public:
  Inst*         inst;
  Block*        block;
  DepEdgeList   succs;
  DepEdgeList   preds;
  PropResource* resource;
  int32_t       weight;
  int32_t       trg_cycle;
  int32_t       dep_cnt;
  bool          visited;
};

//----------------------------------------------------------------------------//

class DepEdge : public TangoObj {
public:
  DepEdge(DepNode* src, DepNode* trg, int32_t latency, int32_t distance, DepType t) :
    src(src), trg(trg), latency(latency), type(t), distance(distance) {}
  DepNode*    src;
  DepNode*    trg;
  int32_t     latency;    // minimum cycle difference
	int32_t     distance;   // iterations difference
  DepType     type;
  bool        isAntiEdge() const { return type == DepTypeAnti && !trg->inst->isMemWrite(); }
  bool        isInterEdge() const { return distance != 0; }
};

//----------------------------------------------------------------------------//

class DepGraph : public TangoObj {
public:
  DepGraph()        { _root_node = NULL; }

  void              init() {
    node_map.clear();
  }

  DepNode*          newDepNode(Inst*, Block*);
  DepNode*          node(Inst* inst) const { return node_map.at(inst); }
  void              newDepEdge(DepNode*, DepNode*, int32_t, DepType=DepTypeUnknown) const;
  void              newDepEdgeFromRoot(DepNode*, int32_t, DepType=DepTypeRoot) const;
  void              removeNode(DepNode*);
  void              removePreds(DepNode*);
  void              newInterEdge(DepNode*, DepNode*, int32_t, int32_t,  DepType=DepTypeUnknown) const; // new inter-iteration edge
  void              setRootNode(DepNode* root_node) { _root_node = root_node; }
  DepNode*          getRootNode() const { return _root_node; }
  Inst2DepNodeMap&  getNodeMap() { return node_map; }


private:
  void              newDepEdgeRaw(DepNode*, DepNode*, int32_t, int32_t, DepType) const;
  DepNode*          _root_node;
  Inst2DepNodeMap   node_map;        // block insts -> DepNode
};

//----------------------------------------------------------------------------//

class DepGraphBuilder : public TangoObj {
public:
  DepGraphBuilder(Ir* ir) : ir(ir),
    use_preds_for_stores(ir->_UsePredsForStores),
    superblock_scheduling(ir->_SuperblockScheduling),
    disable_superblock_scheduling(!ir->_SuperblockScheduling),
    forbid_create_comp_code(ir->_ForbidCreateCompCode),
    use_memory_disambiguation(ir->_UseMemoryDisambiguation),
    clone_helper(ir), cur_loop(nullptr)
  {
    lc_disambiguator  = NULL;
    reg_alloc_helper = NULL;
    st_crossing_locks = 0;
  }
public:
  void                buildSuperblockDeps(Block* block);
  void                buildLoopDeps(Loop* l);
  Inst2DepNodeMap&    depNodeMap(Block* block);

  void                setLcDisamHelper(LockCheckDisambiguator* _lc_disam_helper) { lc_disambiguator = _lc_disam_helper; }
  void                setRegAllocHelper(RegisterAllocHelper* _reg_alloc_helper) { reg_alloc_helper = _reg_alloc_helper; }
  void                setDepGraph(DepGraph* ddg) { dep_graph = ddg; }
  void                setCurLoop(Loop* l) { cur_loop = l; }
  void                setForbidCompCode(bool forbid_comp_code) { disable_superblock_scheduling = forbid_comp_code; }
   // return true if memory address has kind of "register + constant",
  // in this case register and constant operands are passed by arguments.
  bool                getMemoryOperands(Inst* inst, int &reg, int &con);
  // evaluation of cannon representation for base register using input edges for regopnd
  void                evalInDepType(DepNode*, RegOpnd *, Loop* );
  bool                CheckBackEdgeDef(DepNode *dn, RegOpnd *reg, Loop* lp); 
  bool                CheckBackEdgeUse(DepNode *dn, RegOpnd *reg, Loop* lp);
  bool                HasNonInvariantBackEdge(DepNode *dn);
  void                SetNonInarianceForSuccs(DepNode *dn);
  void                SetNonInarianceForPreds(DepNode *dn, RegOpnd* ureg);
  void                SetUsedInUnlinearExpForPreds(DepNode *dn);
  void                SetUnlinearForSuccs(DepNode *dn);
  void                CheckInvarianceForNullCheck(DepNode *dn); 
  bool                CheckInvarianceLoadBeforeNullCheck(DepNode *dn);
  bool                CheckDefIsLinearUsed(DepNode *dn);
  void                evalCannonKind(Inst* inst);
  bool                carryOffsetsCheck(int read_offset, int write_offset, int factor);
  bool                IsBoundCheck(DepNode *dn);
  bool                CheckPredicates(DepNode *dn);
  bool                checkReduction(Inst*);
  bool                exactIntersect(Inst* inst_a, Inst* inst_b);
  int                 checkStoreCanbeLocalized(DepNode *dn, bool createloc);
  void                setMultidefs(DepNode *dn);
  bool                FindRotatingBase(DepNode *dn);
  void                SetAddressEvaluation(DepNode *dn);
  void                fillSplittingClfs(const BlockList&);
  Inst*               FindInduction(RegOpnd *reg);
  Inst*               FindInductionByDep(DepNode* node);
  void                AddCannonKindForInducts();
  bool                FindAntiDep(DepNode* dn, Opnd* opnd);

  const RegOpndList&  getDefBaseAddr() const { return def_base_addr; }
  const RegOpndList&  getUseBaseAddr() const { return use_base_addr; }

private:
  void                calcWeight(Block*);
  void                setDefDataDep(RegOpnd*);
  void                setUseDataDep(Inst*, RegOpnd*);
  void                setTouchUseDataDep(RegOpnd*);
  void                setCondReadDataDep(RegOpnd*);
  void                setPredForStores(Block* block);
  void                resetFields(Block* block);
  void                buildDeps(Block*);

  void                buildNodeMap(Block*);
  void                checkDataDep(Inst*);
  void                checkClfDep(Inst*);
  void                checkMemReadDep(Inst*);
  void                checkMemWriteDep(Inst*);
  void                checkProcedureClfDep(Inst*);
  void                checkSafepointDep(Inst*);
  void                checkAtomicDep(Inst*);
  void                checkSpecialDep(Inst*, InstList&);
  void                checkInlineIntrinsicDep(Inst*);

  // returns false if arrays doesn't intersect
  bool                checkArrayIntersect(Inst* src, Inst* trg);
  // propogation of cannon representation through given instruction
  void                defineInstructionDepType(DepNode*, Inst *, RegOpnd *, Loop *);
  // false if insts definitely different memory
  bool                canMemoryIntersect(Inst* inst_a, Inst* inst_b);
  bool                exactlySameMemory(Inst* inst_a, Inst* inst_b);
  // check if these two store insts are exactly the same in terms of memory disam
  bool                exactlySameStores(Inst* store_a, Inst* store_b);
  Opnd*               getMemoryValueOperand(const Inst* inst) const;  // returns loaded/stored operand
  int32_t             getMemInstLen(const Inst* inst) const;          // returns length of memory inst

  bool                canMoveBeforeClf(const Inst*);
  bool                canMoveAfterClf(const Inst*);
  bool                executeAfterMath(const Inst*);
  bool                checkTheSameArrayIntersect(Inst* src, Inst* trg, Opnd *scon, Opnd *tcon);
  void                addBases(Inst* src, Inst* trg);

  // return true if memory address has kind of "register + constant",
  // in this case register and constant operands are passed by arguments.
  bool                getMemoryOperands(Inst* inst, Opnd **reg, Opnd **con);

  void                buildInterblocksDeps(Block*);
  void                fillPreviousBundles(RegToInstList&, Block*, Block*, int);
  void                processPreviousBundle(RegToInstList&, Block*, Block*, Bundle*, int);
  bool                checkNonlinearDep(RegOpnd*, DepNode *, DepNode *);
  void                delete_reductions(Opnd*);

  bool                isCallSafe(Opnd*);

private:
  Ir*                 ir;
  Block*              block;
  DepGraph*           dep_graph;

  DepNode*            cur_node;
  DepNode*            clf_node;
  DepNode*            prev_clf_node;   // for move stores.
  DepNode*            procedure_clf_node;
  DepNode*            mem_st_wait;
  DepNode*            mem_ld_wait;
  DepNode*            safepoint_node;  // last safepoint node
  DepNode*            atomic_lock_node;// used for adding unlock_st dep
  DepNode*            lsr_def;
  InstList            splitting_clf;
  InstToIntMap        lens_from_root;

  RegOpndList         def_base_addr;   // list of base addreses associated with isMemWrite instructions
  RegOpndList         use_base_addr;   // list of base addreses associated with isMemRead instructions
  RegOpndList         passed_to_call;  // list of base addreses passed to call
  RegOpndToDepNodeMap reg_nodes;

  DepNodeList         rdm_nodes;       // current mem read nodes
  DepNodeList         wrm_nodes;       // current mem write nodes
  DepNodeList         mem_nodes;       // current read & write nodes
  DepNodeList         clf_nodes;
  DepNodeList         procedure_clf_nodes;
  DepNodeList         sfp_nodes;       // inst from last sfp, that def Regs

  // Current maximum number of locks moved over last store
  int32_t             st_crossing_locks;

  bool                use_preds_for_stores;
  bool                forbid_create_comp_code;
  bool                use_memory_disambiguation;
  bool                superblock_scheduling;
  bool                disable_superblock_scheduling;  // temporary disable superblock scheduler.

  Loop*               cur_loop;
  CloneHelper         clone_helper;

  LockCheckDisambiguator* lc_disambiguator;
  RegisterAllocHelper*  reg_alloc_helper;
};

}      // namespace tango
#endif // TANGO INST DEPENDECY
