#ifndef TANGO_INSTRUCTION_SCHEDULER_HPP
#define TANGO_INSTRUCTION_SCHEDULER_HPP

#include "tango/compile_phase.hpp"
#include "tango/clone_helper.hpp"
#include "tango/inst_latency.hpp"
#include "tango/inst_dependencies.hpp"
#include "tango/bundler_helper.hpp"

namespace tango {

//----------------------------------------------------------------------------//

struct  CmpWeight;
struct  CtInfo;
struct  LockCheckInfo;

//----------------------------------------------------------------------------//

typedef arena_multiset <DepNode*, CmpWeight>::type                 DepNodeSortedSet;
typedef arena_map      <Inst*, DepNodeList>::type                  Inst2DepNodeList;
typedef arena_map      <Inst*, InstList>::type                     Inst2InstList;
typedef arena_map      <Inst*, Inst*>::type                        InstToInstMap;
typedef arena_set      <Inst*>::type                               InstSet;
typedef arena_map      <RegOpnd*, CtInfo, RegOpndComparator>::type RegOpnd2CtInfoMap;
typedef arena_list     <RegOpnd*>::type                            RegOpndList;
typedef arena_list     <LockCheckInfo*>::type                      LockCheckInfoList;
typedef arena_map      <Inst*, LockCheckInfo*>::type               Inst2LockCheckInfo;

//----------------------------------------------------------------------------//

#define DPT_UNKNOWN        0
#define DPT_CONSTANT       1
#define DPT_INVARIANT      2
#define DPT_UNDEFINVARIANT 3
#define DPT_UNIFORM        4
#define DPT_LINEAR         5
#define DPT_INDUCTION      6
#define DPT_REDUCTION      7
#define DPT_NONLINEAR      8

//----------------------------------------------------------------------------//

struct CmpWeight {
  bool operator()(DepNode* s1, DepNode* s2) const { return s1->weight > s2->weight; }
};

//----------------------------------------------------------------------------//

class Trace : public BlockList {
public:
  BlockList::iterator   block_iterator;
  BundleList::iterator  bundle_iterator;
  InstList::iterator    inst_iterator;
  InstToIntMap          was_insts; // instructions, that were in this trace and were not in other.
  int                   inst_number;

  Trace() : inst_number(0) {}
};

//----------------------------------------------------------------------------//

struct RestoreSuperblockCfgHelper : public TangoObj {
  RegOpnd2CtInfoMap dispDefOperand;
  Block* curr_target;
  Opnd** curr_label;

  RestoreSuperblockCfgHelper() : curr_target(NULL), curr_label(NULL) {}
};

//----------------------------------------------------------------------------//

struct CtInfo {
  Block* target;
  Opnd** label;
};

//----------------------------------------------------------------------------//

class LockCheckInfo : public TangoObj {
public:
  Inst*       lock;
  Inst*       check;
  OpndVector  use_deps;
  OpndVector  def_deps;
  InstList    dep_insts;
};

//----------------------------------------------------------------------------//

// RegisterAllocHelper is responsible for allocation control and predicate
// registers in sheduler
class RegisterAllocHelper : public TangoObj {
private:

  typedef arena_map<RegOpnd*, int>::type RegCountMap;

  bool        checkFreeCtpr(Inst*);
  bool        checkFreePred(Inst*);
  
  void        updateFreeCtpr(Inst*);
  void        updateFreePred(Inst*);

  // Increments preg_uses_count and free_preds for every pr occurance in vector
  void        incUsedPregs(OpndVector&);

  // Returns allocated pregs count after performing this inst
  int         getFreePredsCount(Inst*);

  // Updates free_preds, preg_uses_count considering OpndVector&
  void        updateAllocatedPredsCount(OpndVector&);
  void        updateAllocatedPredsCount(OpndVector&, int[]);

  // Checks if instr have been moved over some safepoint
  bool        isMovedOverSafepoint(Inst*) const;
  
  // free_ctpr[i] = 0, if i'th ctpr is free, 1 if
  // will be free for next inst, 2 if occupied
  int         free_ctpr[REG_NUM_CTPR];
  // Occurance count of i'th physical preg in
  // uses and cond_reads in current block
  int         free_preds[REG_NUM_PR];
  // preds_delay[i] - how many cycles we need to
  // wait before we can def i'th preg
  int         preds_delay[REG_NUM_PR];
  // preg_uses_count[i] - occurance count for i'th virtual preg
  RegCountMap preg_uses_count;
  // Set of preds that cannot be allocated in scheduler
  OpndSet     unallocatable;

  Inst2DepNodeMap prior_sfps;
  
  Ir*         ir;
  Inst*       prev_sfp;

public:
  RegisterAllocHelper(Ir* ir) : ir(ir), predicate_reg_alloc_flag(ir->_PredicateRegAllocInScheduler) {
    OpndSet defined;
    for (auto block : ir->blocks) {
      for (auto inst : block->insts) {
        for (auto def : inst->unsafeDefs()) {
          if (!def->isPreg()) continue;
          if (defined.count(def)) {
            unallocatable.insert(def);
          } else {
            defined.insert(def);
          }
        }
      }
    }
  }

  // This method adds info about instruction which will be scheduled later
  void        trackLaterInst(Inst*);

  // This method untracks instruction which was supposed to be scheduled
  void        untrackLaterInst(Inst*);

  // This function sets the specified prior safepoint for the node
  void        trackPriorSafepoint(DepNode*, DepNode*);

  // check if there are enough registers to schedule the inst
  bool        checkAvaliableRegs(Inst*);

  // update RegisterAllocHelper according to the scheduled inst
  void        updateAvaliableRegs(Inst*);
  void        init(Block*);
  void        initDelays();
  void        commitBundle(Bundle*);
  void        makePredDelay(int32_t location, int delay) {
    preds_delay[location - EREG_PR0] = std::max(preds_delay[location - EREG_PR0], delay);
  }

  bool        predicate_reg_alloc_flag;
};

//----------------------------------------------------------------------------//

class LockCheckDisambiguator : public TangoObj {
public:
  LockCheckDisambiguator(Ir* _ir) : ir(_ir), clone_helper(_ir) {
    forbid_comp_code = false;
    disable_optimization = false;
  }
  
  void                init();
  void                setBundledInsts(InstSet* _bundled_insts) { bundled_insts = _bundled_insts; }
  void                setDepGraph(DepGraph* _dep_graph) { dep_graph = _dep_graph; }
  
  // building dep graph
  void                addChecksForLoads(Block* block);
  void                setFirstIntersectionStore(Inst* load, Inst* first_intersection);
  void                addIntersectingStore(Inst* load, DepNode* store_node);
  
  // scheduling
  void                lcUpdateAddedInst(Inst* inst, RegisterAllocHelper& alloc_helper);
  void                lcUpdateAddedBundle(Bundle* bundle, RegisterAllocHelper& alloc_helper);
  bool                lcCanAddInst(Inst* inst, Bundle* bundle, int bypass_mask) const;
  DepNode*            tryToDeleteCheck(Inst* lock);
  void                bundleRbranch(Inst* inst, Bundle* bundle, uint32_t bypass_mask) const;

  // restoring
  void                restoreRbranch(Inst* inst, Trace& new_trace, BlockList& comp_blocks);
  void                restoreRbranchJumps();
  void                setForbidCompCode(bool _forbid_comp_code) { forbid_comp_code = _forbid_comp_code; }
  void                setDisableOptimization(bool _disable_optimization) { disable_optimization = _disable_optimization; }
  bool                getDisableOptimization() const { return disable_optimization; }
  
private:
  // building dep graph
  Inst*               createCheckRbranch();
  bool                possibleNullCheck(Inst* inst) const;
  bool                isNullCheckForLoad(Inst* inst, Inst* load) const;
  bool                canBeLockCheck(Inst* load) const; // NB: dep graph must be built before checking
  void                divideLockCheck(Block* block, InstList::iterator inst_it);
  void                restoreLockCheckEdges(DepNode*);  // adds necessary edges when inst can't be divided
  
  // scheduling
  bool                sameLocation(const RegOpnd* reg_a, const RegOpnd* reg_b) const;
  bool                containsReg(const OpndVector& vec, RegOpnd* reg) const;
  bool                isCheckDependent(Inst* inst, LockCheckInfo* lc_info) const;
  bool                canMoveBeforeRbranch(const Inst*, const Inst*) const;
  bool                canAddRbranch(Inst* check, Bundle* bundle, int16_t external_mask) const;
  bool                needCompCode(Inst* check) const;
  
  // restoring
  Block*              addRbranchCompBlock(InstList* comp_insts, Block* original);
  
  Ir*                 ir;
  InstToInstMap       lock_check_map;      // check inst for this lock
  InstSet*            bundled_insts;       // pointer to bundled_insts from scheduler
  InstToInstMap       first_inter_store;   // first std operation which can intersect with lock-check
  Inst2DepNodeList    lock_dep_stores;     // list of store nodes, upon which this store can depend
  Inst2InstList       check_deps;          // list of dependent insts of this check which can be moved over it
  InstToInstMap       check_rbranch_map;
  LockCheckInfoList   opened_locks;        // list of scheduled locks without scheduled checks
  Inst2LockCheckInfo  check_info;
  CloneHelper         clone_helper;
  DepGraph*           dep_graph;           // pointer to dep_graph from scheduler
  BlockList           rbranch_comp_blocks; // used to restore jumps
  bool                forbid_comp_code;
  bool                disable_optimization;
};

//----------------------------------------------------------------------------//

class InstScheduler : public CompilePhase {
public:
  InstScheduler(Ir* ir) :
    CompilePhase(ir, PhaseName::ISR),
    cur_cycle(0),
    superblock_scheduling(ir->_SuperblockScheduling && !ir->_TangoGreedyScheduler),
    clone_helper(ir),  reg_alloc_helper(ir), lc_disambiguator(ir),
    ddg_builder(ir), bundler_helper(ir) {
    lc_disambiguator.setBundledInsts(&bundled_insts);
    lc_disambiguator.setDepGraph(&dep_graph);
    ddg_builder.setLcDisamHelper(&lc_disambiguator);
    ddg_builder.setRegAllocHelper(&reg_alloc_helper);
  }
  Inst2DepNodeMap&    depNodeMap(Block* block);

private:
  // scheduling
  void                processBlock(Block*);
  void                preprocessInstructions(Block*);

  void                schedule(Block*);
  void                pushSuccs(DepNode*, int32_t);
  void                schedule_zero(Block*);
  void                scheduleGreedy(Block*);
  void                fixupLastBundle(Block*);
  bool                canAddToTrace(Block*, int &cmp_num);
  bool                canAddToTraceAfter(Block*);
  bool                checkSpecialRestrictions(Inst*, Bundle*, int32_t) const;
  bool                addInst(DepNode*, Bundle*, int32_t, bool);

  void                clearVisitedMarks(BlockList&);
  void                clearLastDefs();

  void                processBlocks(BlockList&);
  void                processTraceFrom(Block*);

  Trace               chooseTraceFrom(Block*);
  Trace               createTrace(Trace&);
  void                scheduleTrace(Trace&);
  void                repairTrace(Trace&);
  void                restoreInstsFromBundles(Block*);
  void                restoreLiveSets(Trace&);
  void                restoreLayouts(Block*, Block*, Trace&, bool);
  void                restoreJumps(Trace&, Trace&);
  void                mergeBlocks(Trace&);
  Inst*               findNextClfInNewTrace(Trace&, Trace&);
  Inst*               findNextClfInOldTrace(Trace&, Trace&, RestoreSuperblockCfgHelper&);
  void                findTargetForClf(Inst*, RestoreSuperblockCfgHelper&);
  void                processInstructionInTraces(Inst*, Trace&, Trace&);
  void                restoreJump(RestoreSuperblockCfgHelper&, Inst*, Trace&, Trace&);
  void                restoreSwitch(Inst*, Block*, Block*);
  void                restoreMultiCatch(Inst*, Trace&, Trace&);
  void                addBranch(Block*, Block*);
  void                findAndSetLayoutPred(Block*);

  bool                checkIsLastLayoutSuccs(Block*, Block*);

  bool                checkLayoutTransfer(Block*, Block*);

  Block*              addCompCode(Trace& old_trace, Block* target);
  InstList            sortMissedInsts(InstToIntMap&);
  InstList            getExceptionCompCode(Inst*, InstList&);

  void                processDispInOldTrace(RestoreSuperblockCfgHelper&, Inst*);
  void                restoreCompBlocksLayouts(); // Restore comp blocks layouts.
  void                processCompBlocks();        // Schedule comp blocks.

  bool                canContainLabel(Inst*);
  
  void                correctClfHandlers(BlockList&);
  void                correctClfHandlers(Block*);
  void                correctClfHandlers(Block*, InstList&);
  void                fixHardwareInstructions(BlockList&);
  Inst*               nextInstForBundles(Trace&);

protected:
  virtual void        run_internal();

private:
  DepGraph               dep_graph;
  DepGraphBuilder        ddg_builder;
  DepNodeSortedSet       cands;

  Block*                 exit_block;      // block with no successors

  InstSet                bundled_insts;   // set of insts which already distributed to bundles

  BlockList              comp_blocks;     // compensation blocks.

  int32_t                cur_cycle;       // global counter for scheduler.

  bool                   superblock_scheduling;

  RegisterAllocHelper    reg_alloc_helper;
  CloneHelper            clone_helper;
  BundlerHelper          bundler_helper;
  
  LockCheckDisambiguator lc_disambiguator;
};

//----------------------------------------------------------------------------//

}      // namespace tango
#endif // TANGO_INSTRUCTION_SCHEDULER_HPP
