#ifndef TANGO_LOCAL_LOGGER_HPP
#define TANGO_LOCAL_LOGGER_HPP

#include "tango/global_logger.hpp"

#ifdef TANGO_LOG_ENABLED

namespace tango {

class LocalLogger;

extern thread_local LocalLogger* logger;

//-------------------------------------------------------------------------------------//

#define LOG_DEBUG if(logger->isDebug()) *logger

//-------------------------------------------------------------------------------------//

typedef arena_vector<PhaseName>::type PhaseVector;
typedef arena_vector<bool>::type      ContextVector;

enum class LogStage : uint8_t {
  BEFORE,
  AFTER
};

//-------------------------------------------------------------------------------------//

class LocalLogger {
 public:
  LocalLogger();
  
  template <typename T>
  LocalLogger& operator <<(const T);
  
  bool isDebug() const { return _log_debug; }
  
  void init                 (const char*, const char*);
  void start                (PhaseName);
  void end                  ();
  void flush                (const char* = nullptr);
  arena_string getMethodName() { 
    arena_stringstream oss;
    oss << _class_name << "::" << _method_name;
    return oss.str();
  }
  
  //for external explicit use
  void printAsm             (const char* = NULL);
  void printBlocks          (const char* = NULL);
  void printLoops           (const char* = NULL);
  void printCfgDot          (const char*);

 private:
  void log                  (LogStage);
  void putDebugHeader       ();
  void setOption            (PhaseName);
  
  //for implicit use based on command option
  void printAsmInternal     ();
  void printCfgDotInternal  (LogStage)                                      const;
    
  void logBlockHead         (const Block*, arena_stringstream&);
  void printBlock           (const Block*, arena_stringstream&);
  void logLoop              (const Loop*,  arena_stringstream&);

  //methods for exclude use of ir->search(), because ir->search() has side effects
  void fillBlocklist        (ConstBlockList&)                               const;
  void fillBlocklist        (const Block*, ConstBlockSet&, ConstBlockList&) const;

  arena_string generateFilename(LogStage)                                   const;

 private:
  PhaseName           _name;
  
  arena_string        _class_name;
  arena_string        _method_name;

  bool                _log_asm;
  bool                _log_debug;
  bool                _log_graph;
  
  bool                _has_debug;
  bool                _single_thread;

  size_t              _header_size;

  TangoStream&        _out;
  GlobalLogger&       _logger;
  arena_stringstream  _buffer;
  
  ContextVector       _debug_context;
  PhaseVector         _current_phase; //stack based on vector, required for ambiguate nestness
};

//-------------------------------------------------------------------------------------//

template<typename T>
LocalLogger& LocalLogger::operator<<(const T value) {
  if (!_has_debug) {
    putDebugHeader();
    
    if (_single_thread) {
      _out << _buffer.str();
      _buffer.str("");
      _buffer.clear();
    }
  }
  
  if (_single_thread) {
    _out << value;
  } else {
    _buffer << value;
  }

  return *this;
}

//-------------------------------------------------------------------------------------//

}      // namespace tango
#endif // PRODUCT
#endif // SHARE_VM_TANGO_LOCAL_LOGGER_HPP
