#ifndef POST_DOM_TREE
#define POST_DOM_TREE

#include "tango/log.hpp"

namespace tango {

typedef arena_vector<int>::type IntVector;

// This class finds immediate postdominator for block (first successor of block
// such that all paths to exit node goes through it). It is based on algorithm from
// Ron Cytron, Jeanne Ferrante, Barry K. Rosen, Mark N. Wegman, F. Kenneth Zadeck
// Efficiently computing static single assignment form and the control dependence graph
// with few changes:
//
// 1. Common exit block is missing, all blocks with INST_CT_RET are considered as exit blocks
// 2. Because of p.1 there're multiple blocks that dominate itself. In fact, all exit blocks have this property.
//    Due to that sometimes in dominators intersection we can fall into infinite loop. Consider this example:
//            i
//          /   \
//         e1    e2
//
//    Here exit blocks e1 and e2 are postdominators of block i and they have themselves as immediate post dominators.
//    In such cases i is considered to have itself as immediate postdominator.
// 3. BFS used instead of DFS after using ir.search(). The reason for that is invalid loop handling after DFS. Example:
//          ...
//           \
//           11
//          / \\
//         12  10
//         /
//        ...
//   blocks wiil be traversed after DFS in order: ... 12 10 11 ...
//   Algorithm will fail on block 10 cause it has no proceeded successors.
class PostDomTree : public TangoObj {

public:
  PostDomTree(Ir* ir): _ir(ir) {}

  void        build();
  Block*      idom(Block* b) const {
    assert_rvm(b->id >= 0 && b->id < _idx.size(), "bad boundary");
    int idx = _idx[b->id];
    assert_rvm(idx >= 0 && idx < _doms.size(), "bad boundary 2");
    return _doms[idx];
  }
  Ir*         ir() const { return _ir; }

private:
  Ir*         _ir;
  BlockVector _doms;
  IntVector   _idx;

  BlockList   findExitBlocks();
  bool        isExitBlock(Block *block) const;
  void        resetState() const;
  bool        updateIdom(Block *proceeded);
  Block*      intersect(Block *b1, Block *b2) const;
};

} // namespace tango

#endif // POST_DOM_TREE
