#ifndef TANGO_SOFTWARE_PIPELINER_HPP
#define TANGO_SOFTWARE_PIPELINER_HPP

#include "tango/compile_phase.hpp"
#include "tango/inst_dependencies.hpp"
#include "tango/bundler_helper.hpp"
#include "tango/cfg.hpp"
#include <functional>

#ifndef MONO
#define SUPPORT_LOOP_COND
#endif

namespace tango {

//----------------------------------------------------------------------------//

class     BundlerHelper;
class     SMSchedule;
class     PipeliningAllocator;
struct    SPLAssumption;

//----------------------------------------------------------------------------//

typedef arena_map<int32_t, InstList>::type  IntToInstList;
typedef arena_map<RegOpnd*, int32_t>::type  RegToInt;
typedef arena_map<int32_t, RegOpnd*>::type  IntToReg;
typedef arena_set<Opnd*>::type              OpndSet;
typedef arena_set<Inst*>::type              InstSet;
typedef arena_map<DepNode*, int32_t>::type  DepNodeToInt;
#if (__LCC__ > 122)
typedef arena_unordered_set<Opnd*>::type OpndHashSet;
typedef arena_unordered_map<Opnd*, Opnd*>::type OpndToOpnd;
typedef arena_unordered_map<Opnd*, Inst*>::type OpndToInst;
typedef arena_unordered_map<Inst*, Opnd*>::type InstToOpnd;
#else
typedef arena_set<Opnd*>::type OpndHashSet;
typedef arena_map<Opnd*, Opnd*>::type OpndToOpnd;
typedef arena_map<Opnd*, Inst*>::type OpndToInst;
typedef arena_map<Inst*, Opnd*>::type InstToOpnd;
#endif
struct pairhash {
public:
  template <typename T, typename U>
  std::size_t operator()(const std::pair<T, U> &x) const {
    return std::hash<T>()(x.first) ^ std::hash<U>()(x.second);
  }
};
typedef std::pair<Opnd*, uint64_t> OpndInt;
typedef std::pair<Opnd*, Opnd*> OpndOpnd;
#if (__LCC__ > 122)
typedef arena_unordered_map<OpndInt, Opnd*, pairhash>::type OpndIntToOpnd;
typedef arena_unordered_map<OpndOpnd, Opnd*, pairhash>::type OpndOpndToOpnd;
typedef arena_unordered_map<OpndInt, Inst*, pairhash>::type OpndIntToInst;
#else
typedef arena_map<OpndInt, Opnd*>::type OpndIntToOpnd;
typedef arena_map<OpndOpnd, Opnd*>::type OpndOpndToOpnd;
typedef arena_map<OpndInt, Inst*>::type OpndIntToInst;
#endif
typedef arena_vector<SPLAssumption>::type AssumptionsVector;

//----------------------------------------------------------------------------//

const int SPL_MAX_ASM_ATTEMPTS = 10;
const int SPL_MAX_II_ATTEMPTS = 20;
const float SPL_MIN_MOVA_FREQ = 10.0f;
const float SPL_MIN_SPL_FREQ = 4.0f;

//----------------------------------------------------------------------------//

// Helper extension for single-block loop
class LoopExtension : public TangoObj {
public:
  LoopExtension(Ir* ir, Loop* loop) : _ir(ir), _loop(loop) {
    induction_variable = nullptr;
    initial_ind_var = nullptr;
    initial_inst_list = InstList();
  }

  bool                hasPredicateDefs();
  bool                hasExpliticClf();
  bool                hasSideEffect(const Inst* inst);

  bool                usedLater(Opnd*);

  // Checks if register is needed for restoring registers
  bool                usedInRestoration(Opnd*);
  void                fillLiveOutSet();
  const OpndSet&      liveOutSet() const { return _live_out_regs; }

  bool                isLoopInvariant(Opnd*);

  // Allocator helper
  Opnd*               getInitialReg(Opnd*) const;
  Opnd*               getInitialPredicate(Inst* inst) const;

public:
  RegOpnd*            initial_ind_var;
  RegOpnd*            induction_variable;

  // These fields should be moved to special class - allocator helper
public:
  // Used to restore initial regs if pipelining failed
  OpndToOpnd          regs_restore_map;
// info for restoring initial block configuration if pipelining failed
  InstList            initial_inst_list;
// info for restoring list of array access iterators
  InstList  arr_access_iters;

private:
  Ir*                 _ir;
  Loop*               _loop;
  OpndSet             _live_out_regs;

};

// Here is used Swing Modulo Scheduler algorithm. Check
// llvm's MachinePipeliner.cpp file for detailed exlplanation

// This class repesents the scheduled code.  The main data structure is a
// map from scheduled cycle to instructions.  During scheduling, the
// data structure explicitly represents all stages/iterations.   When
// the algorithm finshes, the schedule is collapsed into a single stage,
// which represents instructions from different loop iterations.
//
// The SMS algorithm allows negative values for cycles, so the first cycle
// in the schedule is the smallest cycle value.
class SMSchedule : public TangoObj {
public:
  SMSchedule(Ir* ir, LoopExtension* loop_ext) : _ir(ir), _loop_ext(loop_ext),
    _bundler_helper(ir) {}

  void          reset();

  void          setInitiationInterval(int ii) { _initiation_interval = ii; }
  int           getFirstCycle() const { return _first_cycle; }
  int           getLastCycle() const { return _last_cycle; }
  int           getFinalCycle() const { return _first_cycle + _initiation_interval - 1; }
  int           stageScheduled(Inst* inst) const {
    if (!_inst_to_cycle.count(inst)) {
      return -1;
    }
    return (_inst_to_cycle.at(inst) - _first_cycle) / _initiation_interval;
  }
  BundleList    getFinalSchedule() const { return _final_schedule; }
  int32_t       getMaxStageCount() const { return (_last_cycle - _first_cycle) / _initiation_interval; }
  InstList      getScheduledInsts(int cycle) const {
    if (!_scheduled_insts.count(cycle)) return InstList();
    return _scheduled_insts.at(cycle);
  }

  SPLAssumption computeStart(DepNode* cur_node, int32_t& max_early_start,
                             int32_t& min_late_start, int32_t& min_end,
                             int32_t& max_start, const DepGraph& dep_graph);
  // Estimates max possible stage based on  scheduling and remaining insts
  // It is needed used when scheduling side-effect operation which are need
  // to be placed on last stage
  int           estimateMaxStage(DepNodeList::iterator ni,
                                 DepNodeList::iterator ne,
                                 const DepGraph& dep_graph);
  bool          scheduleInst(DepNode* cur_node, int32_t start_cycle,
                             int32_t end_cycle);
  bool          addInst(Bundle* bundle, Inst* inst, int32_t cur_cycle,
                        int32_t external_mask) const;
  void          addBundle(BundleList& bundles, Bundle* bundle);
  void          movasToFirstStage(DepGraph& dep_graph, Loop* loop);
  void          finalizeSchedule(DepGraph& dep_graph, Loop* loop, Inst* loop_ct_inst);
  void          restoreRegs(Loop* loop);
  void          replaceInitialRegs(Loop* loop);

  void          setPrepBlock(Block* prep) { _prep_block = prep; }

  void          logFlatSchedule();
  void          logFinalSchedule();

private:
  Ir*            _ir;
  LoopExtension* _loop_ext;
  BundlerHelper  _bundler_helper;
  // map from execution cycle to instructions.
  IntToInstList  _scheduled_insts;

  // map from instruction to execution cycle.
  InstToIntMap   _inst_to_cycle;

  // keep track of the first cycle value in the schedule. It starts
  // as zero, but the algorithm allows negative values.
  int32_t        _first_cycle;

  // keep track of the last cycle value in the schedule.
  int32_t        _last_cycle;

  // the initiation interval (II) for the schedule.
  int32_t        _initiation_interval;

  // schedule after finalizing (all insts in first stage)
  BundleList     _final_schedule;

  Block*         _prep_block;
};

//----------------------------------------------------------------------------//

// This helper class provides some easy peephole optimizations
// on the one-block loop
class PeepholeOptimizer : public TangoObj {
public:
  PeepholeOptimizer(Ir* ir, Loop* loop, LoopExtension* loop_ext,
    DepGraphBuilder* dep_builder) : _ir(ir), _loop(loop),
    _loop_ext(loop_ext), _dep_builder(dep_builder) {}

  void run() {
    coalesceSxt();
    coalesceShl();
    coalesceAdds();
    propagateLoadImms();
  }

private:
  // This function coalesce some instruction into one. This is important for
  // non-optimized loops like ones in Mono.
  // The basic conditions for an instruction if it can ba coalesced (like it's
  // opcode) are checked by prechecked function.
  // Then the special map with arguments of this instructions is defined where
  // key is user-defined as template. All instructions with the same set the
  // same key are coalesced.
#if (__LCC__ > 122)
  template <typename Key, typename KeyHash = std::hash<Key>,
            typename MapType = typename arena_unordered_map<Key, Opnd*, KeyHash>::type>
  void coalesceInstructions(std::function<bool(Inst*)>& precheck,
                            std::function<Key(Inst*)>& extractKey) {
#else
  template <typename Key, typename MapType = typename arena_map<Key, Opnd*>::type>
  void coalesceInstructions(std::function<bool(Inst*)>& precheck,
                            std::function<Key(Inst*)>& extractKey) {
#endif
    MapType map;
    Block* block = _loop->head;
    auto next_it = block->insts.begin();
    OpndToOpnd uses_replacement;
    auto updateUses = [&](OpndVector& uses) {
      for (auto use_it = uses.begin(); use_it != uses.end(); ++use_it) {
        Opnd* use = *use_it;
        if (uses_replacement.count(use)) {
          *use_it = uses_replacement[use];
        }
      }
    };

    for (auto inst : block->insts) {
      updateUses(inst->uses);
      updateUses(inst->touch_uses);

      // Check if this instruction is a candidate for coalescing
      if (!precheck(inst)) continue;

      for (auto def : inst->defs()) {
        if (_loop_ext->usedLater(def)) continue;
        if (def->asReg()->isAllocated()) continue;
        auto key = extractKey(inst);
        if (map.count(key)) {
          uses_replacement[def] = map[key];
        } else {
          map[key] = def;
        }
      }
    }
  }

  void coalesceSxt();
  void coalesceShl();
  void coalesceAdds();
  void propagateLoadImms();

private:
  Ir*            _ir;
  LoopExtension* _loop_ext;
  Loop*          _loop;

  // used to get memory operands TODO: move this function out of the class
  DepGraphBuilder* _dep_builder;
};

//----------------------------------------------------------------------------//

// This helper class which allocates loop registers
class PipeliningAllocator : public TangoObj {
public:
  PipeliningAllocator(Ir* ir, LoopExtension* loop_ext) : _ir(ir),
    _loop_ext(loop_ext) {}
  // Registers allocation
  bool            allocateRegisters(Loop* loop, SMSchedule& schedule);
  // Allocates all uses and defs for inst in loop. Rewrites max used breg
  bool            allocateInstsRegisters(Loop* loop, Inst* inst,
                                         int* max_odd, int* max_even,
                                         SMSchedule& schedule);
  bool            allocateAntiDep(Loop* loop, RegOpnd* alloc_reg,
                                  int* max_reg, SMSchedule&);
  bool            allocateFlowDep(Loop* loop, RegOpnd* alloc_reg,
                                  int* max_reg, SMSchedule& schedule);
  bool            replaceReg(OpndVector::iterator reg_it, int32_t location = UNALLOCATED);

  int32_t         getBregsCount() const    { return _bregs_cnt; }
  int32_t         getBPredsCount() const   { return _bpreds_cnt; }
  const RegToInt& getRegsInitiation() const{ return _regs_initiation; }

private:
  Ir*            _ir;
  LoopExtension* _loop_ext;
  int32_t        _bregs_cnt;
  int32_t        _bpreds_cnt;

  // Map from initial registers to bregs numbers where they
  // need to be allocated in prep block
  RegToInt       _regs_initiation;
};

//----------------------------------------------------------------------------//

// In some cases we want to retry our pipelining with the same II but some
// additional assumtions. This struct contains assumption about the inst
struct SPLAssumption : public TangoObj {
  Inst* inst;
  int32_t early_start;
  int32_t late_start;

  SPLAssumption() : inst(nullptr) {}
  SPLAssumption(Inst* inst, int32_t early_start, int32_t late_start) :
    inst(inst), early_start(early_start), late_start(late_start) {}
};

//----------------------------------------------------------------------------//

class SoftwarePipeliner : public CompilePhase {
public:
  SoftwarePipeliner(Ir* ir) : CompilePhase(ir, PhaseName::SPL), _bundler_helper(ir),
    _clone_helper(ir),_ddg_builder(ir) {}


protected:
  virtual void         run_internal();
private:
  // Preparing
  void                 pipelineDiver(Loop* _loop, int* loops_pipelined);
  bool                 canBePipelined(Block* block);
  bool                 checkAntiDefs(Block* block) const;
  bool                 checkBlockIsLoop(Loop* loop);
  void                 saveInitialInsts(Block* block);
  bool                 pipelineBlock(Block* block);
  void                 fillLiveOutSet();
  void                 init(Block* prologue);
  void                 buildLoopDeps();

  // After if-conversion we can get a lot of operation under predicates which
  // can cause many dependencies, though, due to using bregs most of these
  // dependencies may be redundant. This function removes unnecessary predicates
  void                 preparePredicates();
  void                 fixAntiDeps();
  void                 deleteCtInst();

  // Layouting
  void                 createCondBranch();
  void                 createCondBranch(Block* cond_trg, Inst* ct_cond);

  // Ordering and transforming to mova
  void                 buildNodeOrder();
  void                 removeRedundantLoads();
  void                 removeRedundantInstructions();
  void                 useHardwareInstructions();
  bool                 canUseMovaInstructions();
  int32_t              estimateMII() const;

  // Pipelining
  bool                 pipeline(SMSchedule& schedule,
                                PipeliningAllocator& pipelining_allocator);
  bool                 tryModuloSchedule(SMSchedule&, int32_t);
  bool                 checkFinalSchedule(SMSchedule& schedule);

  // Forbids bregs allocations for invariants and regs in live out set
  void                 forbidAllocateBregs(PipeliningAllocator&
                                           pipelining_allocator, Loop* loop);

  // Generating prologue and epilogue
  Block*               generatePrepareBlock(const PipeliningAllocator& alloc,
                                            const SMSchedule& schedule,
                                            Block* prep_block);
  bool                 patchMovaAls(BundleList& bundles, int32_t movas[]);
  InstList             generateLdispParametres(Inst* mova, Opnd** aad, Opnd** aaind,
                                              Opnd** aaincr, Opnd** fapb, IntToReg& incr_map,
                                              int32_t cur_asz[], int32_t cur_area[], int32_t movas[]);

  // Generate prepare operations for using mova
  void                 generateHardwarePrepare(const SMSchedule& schedule,
                                               Block* prep_block);

  Block*               generatePrologue(SMSchedule& schedule,
                                        Block* prep_block);
  Block*               generateEpilogue(SMSchedule& schedule,
                                        Block* finish_block);
  Block*               generateFinishBlock(Block* finish_block);
  void                 genPseudoDefsForBregs();
  void                 generateHardwareFinish(Block* finish_block);

  // We need to generate a non-pipelined version of loop for small iterations cases
  void                 generateLoopCopy();
  InstList             fillLsrReg();
  void                 calculateIterationsCount();
  // Registers restoration used with implicit epilogue
  void                 restoreLiveOutRegs(const SMSchedule& schedule,
                                          Block* epilogue);
  BundleList           restoreLiveOutInst(const SMSchedule& schedule,
                                          Inst* inst);
  void                 addCheckBlockToCFG(SMSchedule& schedule);
  void                 restoreInitialBlock(Block* block);

  // dep_graph
  void                 computeNodeFunctions();
  int32_t              getASAP(DepNode* node)   const {return _asap_map.at(node); }
  int32_t              getALAP(DepNode* node)   const { return _alap_map.at(node); }
  int32_t              getMOV(DepNode* node)    const { return getALAP(node) - getASAP(node); }
  int32_t              getDepth(DepNode* node)  const { return _depth_map.at(node); }
  int32_t              getHeight(DepNode* node) const { return _height_map.at(node); }

  // Helper functions
  bool                 hasImplicitEpilogue()    const { return _loop_ext->hasExpliticClf(); }
  bool                 do_pipeline(Loop* loop, int* loops_pipelined);
private:
  PipeliningAllocator* _pipelining_allocator;
  LoopExtension*       _loop_ext;
  CloneHelper          _clone_helper;
  Loop*                _loop;
  DepGraph             _dep_graph;
  DepNodeList          _nodes_order;
  BundlerHelper        _bundler_helper;
  DepGraphBuilder      _ddg_builder;

  Inst*                _loop_ct_inst;
  Inst*                _loop_disp_inst;
  Inst*                _loop_setwd_inst;
  Block*               _cloned_block;
  Block*               _check_block;

  RegOpnd*             _iterations_count;
  DepNodeToInt         _asap_map;
  DepNodeToInt         _alap_map;
  DepNodeToInt         _depth_map;
  DepNodeToInt         _height_map;

  int                  _store_cnt_in_meth;
  int                  _load_cnt_in_meth;
  int                  _store_cnt;
  int                  _load_cnt;
  AssumptionsVector    _assumptions;
  InstList             _alive_inducts;
  bool                 _has_mova;
};

//----------------------------------------------------------------------------//

}       // namespace tango
#endif  // TANGO SOFTWARE PIPELINER HPP
