#ifndef TANGO_STACK_HPP
#define TANGO_STACK_HPP

#include "tango/tango_inst.hpp"
#include "tango/utilities/bitMaps.hpp"

namespace tango {

//----------------------------------------------------------------------------//
//   1) stack grows down
//   2) loc_size = in_arg + locals
//   3) if (loc_size < 4) loc_size = 4 (in quadro regs)
//   4) if (out_size < 4) out_size = 4 (in quadro regs)
//
//   +================================+ <--- frame start
//   | in arg 0             (8 bytes) |
//   | in arg 1             (8 bytes) |
//   | ...                            |
//   +--------------------------------+
//   |                                |
//   | locals                         |
//   |                                |
//   +--------------------------------+ <--- loc_size
//   | stack pointer        (8 bytes) |
//   +--------------------------------+
//   | out arg 0            (8 bytes) |
//   | out arg 1            (8 bytes) |
//   | ...                            |
//   | out arg 7            (8 bytes) |
//   +================================+ <--- frame end

class RegStack {
public:
  RegStack();
  int32_t   newInArg(int32_t);
  int32_t   newOutArg(int32_t);
  int32_t   getLocSize() const;

  int32_t   newReg(BitMap&, RegType, bool);
  BitMap&   getLegalMask(RegType, bool);

public:
  int32_t   loc_size;
  int32_t   out_size;

  BitMap    preserv_mask_r;
  BitMap    preserv_mask_p;
  BitMap    preserv_mask_c;
  BitMap    preserv_mask_aad;
  BitMap    preserv_mask_aasti;
  BitMap    preserv_mask_aaind;
  BitMap    preserv_mask_aaincr;
  BitMap    preserv_mask_aaldi;
  BitMap    preserv_mask_aalda;
  BitMap    preserv_mask_aaspec;
  BitMap    preserv_mask_s;
  BitMap    scratch_mask_r;
  BitMap    scratch_mask_p;
  BitMap    scratch_mask_c;
  BitMap    scratch_mask_aad;
  BitMap    scratch_mask_aasti;
  BitMap    scratch_mask_aaind;
  BitMap    scratch_mask_aaincr;
  BitMap    scratch_mask_aaldi;
  BitMap    scratch_mask_aalda;
  BitMap    scratch_mask_aaspec;
  BitMap    scratch_mask_s;
};

//----------------------------------------------------------------------------//
// stack_pointer = previous_stack_pointer - stack_size (stack grows down)
//
//   +--------------------------------+
//   | ...                            |
//   | in arg 9             (8 bytes) |
//   | in arg 8             (8 bytes) |
//   +--------------------------------+
//   | reserved area       (64 bytes) |
//   +================================+ <--- previous stack pointer
//   |                                |
//   | locals (spill area)            |
//   |                                |
//   +--------------------------------+ <--- local base
//   | ...                            |
//   | out arg 9            (8 bytes) |
//   | out arg 8            (8 bytes) |
//   +--------------------------------+ <--- out args base
//   | reserved area       (64 bytes) |
//   +================================+ <--- stack pointer

#define RESERVED_AREA 64

class MemStack {
public:
  MemStack() : size(0) {}
  int32_t newLocSlot();
  int32_t slotsCount();

public:
  int32_t size;
};

} // namespace tango

#endif // SHARE_VM_TANGO_STACK_HPP
