#ifndef TANGO_ALLOCATOR_H
#define TANGO_ALLOCATOR_H

#include <stdint.h>
#include <stdlib.h>
#include <algorithm>

#include "tango/tango_globals.hpp"


namespace tango {

template<class T>
struct Allocatable {
  typedef T Allocator;
  inline void *operator new(size_t nbytes, T &allocator);
  inline void *operator new(size_t nbytes, T *allocator);
  inline void *operator new(size_t nbytes);
};

// allocator interface
class ITangoAllocator {
public:
  virtual void *allocate(size_t bytes) = 0;
  virtual void release() = 0;
};

// default allocator for tango
class TangoBufferAllocator {
private:
  static constexpr size_t DEFAULT_INITIAL_SIZE = 2 * 1024 * 1024;
  static constexpr float GROWTH_FACTOR = 1.5f;

  struct Chunk {
    uint8_t *buffer;
    uint8_t *current;
    size_t buffer_size;
    Chunk* next;
  };

  Chunk *pool;
  size_t total_size;

  void grow_buffer(size_t min_bytes) {
    auto new_buffer_size = std::max(min_bytes, static_cast<size_t>(total_size * (GROWTH_FACTOR - 1)));
    Chunk *new_chunk = alloc_chunk(new_buffer_size);
    new_chunk->next = pool;
    pool = new_chunk;
    total_size += new_buffer_size;
  }

  Chunk *alloc_chunk(size_t buffer_size) {
    auto *new_chunk = static_cast<Chunk*>(malloc(sizeof(Chunk) + buffer_size));
#ifdef DEBUG
    memset(new_chunk, 0x00, sizeof(Chunk) + buffer_size);
#endif
    new_chunk->buffer = reinterpret_cast<uint8_t*>(new_chunk) + sizeof(Chunk);
    new_chunk->current = new_chunk->buffer;
    new_chunk->buffer_size = buffer_size;
    new_chunk->next = nullptr;

    return new_chunk;
  }

public:
  TangoBufferAllocator(size_t initial_size = 0) {
    size_t initial_buffer_size = DEFAULT_INITIAL_SIZE;
    if (initial_size != 0) {
      initial_buffer_size = initial_size;
    }
    total_size = initial_buffer_size;
    pool = alloc_chunk(initial_buffer_size);
  }

  ~TangoBufferAllocator() {
    release();
  }

  inline void *allocate(size_t bytes) {
    assert_rvm(pool, "pool == 0");
    
    if (!is_size_aligned(bytes, 2 * BytesPerWord)) {
      bytes = align_size_up(bytes, 2 * BytesPerWord);
    }
    if (pool->current + bytes >= pool->buffer + pool->buffer_size) {
      grow_buffer(bytes);
    }
    void *ptr = pool->current;
    pool->current += bytes;

    return ptr;
  }

  void release() {
    for(Chunk *chunk = pool; chunk != nullptr;) {
      Chunk *next = chunk->next;
#ifdef DEBUG
      memset(chunk, 0x00, sizeof(Chunk) + chunk->buffer_size);
#endif
      free(chunk);
      chunk = next;
    }
  }

  static TangoBufferAllocator& getPersistentAllocator() {
    static TangoBufferAllocator alloc;

    return alloc;
  }
};

template <class Allocator>
void* Allocatable<Allocator>::operator new(size_t nbytes, Allocator &allocator) {
  return allocator.allocate(nbytes);
}

template <class Allocator>
void* Allocatable<Allocator>::operator new(size_t nbytes, Allocator *allocator) {
  return allocator->allocate(nbytes);
}

#ifdef USE_TANGO_INTERFACE_ALLOCATOR
#define TANGO_ALLOCATOR ITangoAllocator
#endif

#ifndef TANGO_ALLOCATOR
#define TANGO_ALLOCATOR TangoBufferAllocator
#endif

typedef Allocatable<TANGO_ALLOCATOR> TangoObj;

extern thread_local TangoObj::Allocator *_shared_allocator;

template <class Allocator>
void* Allocatable<Allocator>::operator new(size_t nbytes) {
  return _shared_allocator->allocate(nbytes);
}

typedef Allocatable<TangoBufferAllocator> PersistentObj;

struct AllocatorUtils : public TangoObj {
  template <int alignment, typename U>
  inline static U allocate_aligned(Allocator *a, size_t bytes) {
    uintptr_t res = reinterpret_cast<uintptr_t>(a->allocate(bytes + alignment));
    return reinterpret_cast<U>((res + alignment) & ~alignment);
  }
};


} // tango
#endif //TANGO_ALLOCATOR_H
