#ifndef TANGO_TREE_BALANCER_HPP
#define TANGO_TREE_BALANCER_HPP

#include "tango/compile_phase.hpp"

namespace tango {

//----------------------------------------------------------------------------//

class TreeBalancer : public CompilePhase {
  struct Tree {
    int32_t root_index;
    int32_t left_index;
    int32_t right_index;

    int32_t left_height;
    int32_t right_height;

    int32_t valid_in_block;
    int32_t used_in_block;

    int32_t position;

    bool left() const {
      return left_height > right_height;
    }

    int32_t height() const {
      return highHeight() + 1;
    }

    int32_t high() const {
      return left() ? left_index : right_index;
    }

    int32_t highHeight() const {
      return std::max(left_height, right_height);
    }

    int32_t highIndex() const {
      return left() ? 1 : 2;
    }

    int32_t lowHeight() const {
      return std::min(left_height, right_height);
    }

    void set(int32_t lindex, int32_t lheight, int32_t rindex, int32_t rheight) {
      left_index = lindex;
      left_height = lheight;

      right_index = rindex;
      right_height = rheight;
    }

    void setLeft(const Tree& t) {
      left_index = t.root_index;
      left_height = t.height();
    };

    void setRight(const Tree& t) {
      right_index = t.root_index;
      right_height = t.height();
    }

    void init(int32_t index, int32_t block) {
      root_index = index;
      position = index;
      left_height = 0;
      left_index = -1;
      right_height = 0;
      right_index = -1;
      valid_in_block = block;
      used_in_block = -1;
    }

    Tree() :
      valid_in_block(-1),
      used_in_block(-1) {}
  };

public:
  TreeBalancer(Ir* ir) : CompilePhase(ir, PhaseName::ATB) {
  }
  
private:
  void moveInstruction(int32_t inst_index, int32_t dest_index);
  void swapInputs(Tree& t);
  void swapOperation(Tree& t);

  bool valid(int32_t inst_index) const;
  bool oneOut(int32_t inst_index) const;
  int32_t precedence(int32_t inst_index) const;
  
  void processBlock();
  void countOuts() const;
  void balance(Tree& root, Tree& high, bool left1, bool left2);
  void balance(int32_t root_index, Tree& t, bool left);
  void balance(int32_t root_index, Tree& l, Tree& r);

  Tree newNode(int32_t root_index, Tree& t, bool left);
  Tree newNode(int32_t root_index, Tree& hi, Tree& lo, bool left);

  void tryBalance(int32_t index, Tree& t, bool left);
  void tryBalance(int32_t index, Tree& l, Tree& r);

  bool canBalance(int32_t index, const Tree& t) const;
  bool canBalance(int32_t index, const Tree& l, const Tree& r) const;
  bool redef(const Tree& t) const;

  Opnd* getDef(int32_t index) const {
    return (*iterator_table[index])->getDef(0);
  }

  bool validRes(int32_t inst_index) const {
    return tree[getDef(inst_index)->id].root_index == inst_index;
  }


protected:
  void run_internal() override;

private:
  arena_vector<InstList::iterator>::type iterator_table;
  arena_vector<Tree>::type tree;

  Block* current_block;
  
  mutable arena_vector<int32_t>::type out_count;
  mutable bool outs_counted;
};

//----------------------------------------------------------------------------//

} // namespace tango
#endif // TANGO_TREE_BALANCER_HPP